import json
import string
from datetime import datetime, date
from auditlog.models import PersistentLog
from rest import helpers as rest_helpers
from rest.uberdict import UberDict
from rest.encryption import DECRYPTER
from rest import ua
from objict import objict

from django.http.request import QueryDict

NOTFOUND = "311!@#$%^&*"
SAFE_ASCII = set(string.printable)
# load decrypter


def createFakeRequest(method="GET", **kwargs):
    member = kwargs.get("member", objict(is_authenticated=True, is_staff=True, is_superuser=True))
    return objict(
        META={}, 
        user=member,
        member=member,
        group=kwargs.get("group", None),
        method=method, 
        _started=0, 
        DATA=RequestData(data=kwargs))


class RequestData(object):

    @staticmethod
    def upgradeRequest(request, data=None):
        if not hasattr(request, "DATA"):
            request.DATA = RequestData(request, data)
            request.DATA.parse()

    def __init__(self, request=None, data=None):
        self.__data = data
        self.is_logged = False
        self.request = request
        self.ua_info = None
        self.url_params = None
        if not self.request:
            return
        self.request.is_sns = False
        self.parsePARAMS()
        self.request.is_json = self.isJSON()

    def parse(self):
        if self.request.is_json and self.request.body:
            self.parseJSON()
        else:
            self.__data = self.params
        if self.__data is None:
            self.__data = {}

    def decrypt(self, key="emsg", nkey=None):
        """
        decrypt any emsg data that may be in this returning the data
        nkey is the fallback unencrypted key to look for
        """
        edata = self.get(key, None)
        if edata is None:
            if nkey:
                return self.get(nkey, None)
            return None
        if not DECRYPTER:
            raise Exception("decryption key not loaded")
        data = DECRYPTER.decrypt(edata)
        try:
            if data and data.startswith('{') or data.startswith('['):
                # data = json.loads(data)
                return UberDict.fromJSON(data)
        except Exception:
            pass
        return data

    def parsePARAMS(self):
        if self.request.method == "POST":
            data = self.request.POST
            if self.request.GET:
                self.url_params = UberDict.fromdict(self.normalizeData(self.request.GET))
        elif self.request.method == "GET":
            data = self.request.GET
        elif self.request.method == "DELETE":
            data = dict()
        elif self.request.method == "PUT":
            if self.request.content_type == "application/json":
                data = UberDict.fromJSON(self.request.body, ignore_errors=True)
            else:
                if hasattr(self.request, '_post'):
                    del self.request._post
                    del self.request._files
                try:
                    self.request.method = "POST"
                    self.request._load_post_and_files()
                    self.request.method = "PUT"
                except AttributeError:
                    self.request.META['REQUEST_METHOD'] = 'POST'
                    self.request._load_post_and_files()
                    self.request.META['REQUEST_METHOD'] = 'PUT'
                data = self.request.POST
        else:
            data = dict()
        self.params = UberDict.fromdict(self.normalizeData(data))
        if self.url_params:
            self.params.extend(self.url_params)

    def parseJSON(self):
        try:
            self.__data = UberDict.fromJSON(self.request.body, ignore_errors=False)
            if isinstance(self.__data, list):
                self.__data = UberDict(data=self.__data)
            self.__data.update(self.params)
            # data = UberDict.fromdict(dict(self.params))
            # self.__data = self.normalizeData(json.loads(self.request.body), data)
        except Exception:
            rest_helpers.log_exception(
                "request error", "{} -> {} {}".format(self.request.ip, self.request.method, self.request.path),
                dict(self.request.META))
            rest_helpers.log_error("request body", self.request.body)
            self.request.is_json = False
            return

    def isJSON(self):
        if "HTTP_X_AMZ_SNS_MESSAGE_TYPE" in self.request.META:
            self.request.is_sns = True
            return True
        ct = self.request.META.get('CONTENT_TYPE')
        if self.request.method == "POST" and ct:
            return "json" in ct.lower()
        return False

    def getHeader(self, name):
        if type(name) is list:
            for n in name:
                v = self.getHeader(n)
                if v is not None:
                    return v
        # X-Gitlab-Token
        name = name.upper()
        for key in self.request.META:
            if key.upper() == name:
                return self.request.META.get(key)
        return None

    def remove(self, key, default=None):
        if type(key) is list:
            v = None
            for k in key:
                v = self.remove(k, NOTFOUND)
                if v != NOTFOUND:
                    return v
            return default

        data, pk = self.getDictAndKey(key)
        if data:
            value = data[pk]
            del data[pk]
            return value
        return default

    def getDictAndKey(self, key, data=None):
        """
        This will return the dict that holds the key.
        used when going into nested objects
        """
        if data is None:
            data = self.__data
            if not data:
                return None, None

        if "." in key:
            keys = key.split('.')
            obj = None
            pk = None
            for k in keys:
                if not data:
                    return None, None

                obj, pk = self.getDataAndKey(k, data)
                if obj is None:
                    return None, None

                data = obj.get(key)
            return obj, pk

        for k in data:
            if type(k) in [str, str] and k.lower() == key.lower():
                return data, k
        return None, None

    def getIgnore(self, key, data=None):
        if data is None:
            data = self.__data

        if data:
            for k in data:
                if type(k) in [str, str] and k.lower() == key.lower():
                    return data.get(k)
        return NOTFOUND

    def getNormal(self, key, data=None):
        if data is None:
            data = self.__data

        if key in data:
            return data.get(key)
        return NOTFOUND

    def lowerKeys(self):
        if self.__data:
            self.__data = rest_helpers.lowerKeys(self.__data)

    def removeUnderscores(self):
        if self.__data:
            self.__data = rest_helpers.removeUnderscoreKeys(self.__data)

    def hasRequired(self, keys):
        # this method returns true if it has all the required keys
        for k in keys:
            if k not in self:
                return False
        return True

    def has_key(self, key):
        if type(key) is list:
            for k in key:
                if k in self:
                    return True
            return False
        value = self.get(key, NOTFOUND)
        return value != NOTFOUND

    def __iter__(self):
        return self.__data.__iter__()

    def __getattr__(self, key):
        return self.get(key)

    def __getitem__(self, key):
        return self.get(key)

    def __setitem__(self, key, value):
        return self.set(key, value)

    def __contains__(self, key):
        value = self.get(key, NOTFOUND)
        return value != NOTFOUND

    def __unicode__(self):
        if self.__data:
            return str(self.__data)
        return ""

    def log(self, include_headers=True):
        if not self.is_logged:
            self.is_logged = True
            sanitized = UberDict.fromdict(self.toDict())
            if include_headers:
                headers = {}
                for key in self.request.META:
                    if key in ["HTTP_AUTHORIZATION"]:
                        value = self.request.META.get(key)
                        if value and len(value) > 6:
                            headers[key.upper()] = "{}***{}".format(value[:value.find(' ')], value[-6:])
                sanitized["_headers_"] = headers
            if self.request:
                try:
                    self.request.request_log = PersistentLog.log(sanitized, level=5, component="rest", action="request", request=self.request)
                except Exception as err:
                    print(("PersistentLog.log failed with: '{}".format(str(err))))
                    print((repr(sanitized)))

    def getStrictType(self, value, field_type, default=None):
        if type(value) is field_type:
            return value

        try:
            if value == '':
                if field_type in [int, float]:
                    return field_type(0)

            if field_type in [int, str, float, str]:
                value = field_type(value)
            elif field_type is datetime:
                value = rest_helpers.parseDateTime(value)
            elif field_type is date:
                value = rest_helpers.parseDate(value, as_date=True)
            elif field_type is bool:
                return value in [1, '1', 'y', 'Y', 'on', 'T', 't', 'True', 'true']
            elif field_type is list:
                if value and value != NOTFOUND:
                    if type(value) in [dict, list]:
                        return value
                    elif "," in value:
                        return value.split(',')
                    return [value]
                return []
            elif field_type is dict and isinstance(value, str):
                return objict.fromJSON(value)
        except Exception:
            return default
        return value

    def getList(self, key, default=None, split_spaces=False):
        return self.getlist(key, default, split_spaces)

    def getlist(self, key, default=None, split_spaces=False):
        res = self.get(key, default)
        if res and res != NOTFOUND:
            if type(res) in [dict, list]:
                return res
            elif "," in res:
                resp = res.split(',')
            elif split_spaces and " " in res:
                return res.split(' ')
            return [res]
        return []

    def _recursive_as_dict(self, root, key, value):
        gname = key[:key.find('.')]
        fname = key[key.find('.')+1:]
        if gname not in root:
            root[gname] = dict()
        nroot = root[gname]
        if '.' in fname:
            self._recursive_as_dict(nroot, fname, value)
        elif fname in nroot:
            if not isinstance(nroot[fname], list):
                nroot[fname] = [nroot[fname]]
            nroot[fname].append(value)
        else:
            nroot[fname] = value

    def normalizeData(self, data, normed=None):
        # this method takes "field1.field2 = value" to proper expaned dict
        if normed is None:
            normed = dict()
        is_qdict = isinstance(data, QueryDict)
        for field in data:
            if is_qdict:
                value = data.getlist(field)
                if len(value) <= 1:
                    value = data[field]
            else:
                value = data.get(field)
            if field.endswith('[]'):
                field = field[:-2]
            if "." in field:
                self._recursive_as_dict(normed, field, value)
            elif field in normed:
                if not isinstance(normed[field], list):
                    normed[field] = [normed[field]]
                normed[field].append(value)
            else:
                normed[field] = value
        return normed

    def toDict(self):
        self.__data = self.asDict()
        return self.__data

    def asDict(self, copy=False):
        if self.__data:
            if copy:
                return copy.deepcopy(self.__data)
            return self.__data
        data = None
        if self.request.method == "POST":
            data = self.request.POST
        else:
            data = self.request.GET

        d = self.normalizeData(data)

        if copy:
            return copy.deepcopy(d)
        return d

    def asUberDict(self):
        return UberDict.fromdict(self.__data)

    def asObjict(self):
        return objict.fromdict(self.__data)

    def toObject(self):
        return objict.fromdict(self.__data)

    def set(self, key, value):
        keys = key.split('.')
        data = self.__data
        depth = len(keys)
        for k in keys:
            depth -= 1
            if depth > 0:
                res = self.get(k, data=data)
                if res is None:
                    data[k] = {}
                    data = data[k]
                else:
                    data = res
            else:
                data[k] = value

    def removeNonASCII(self, value):
        return "".join([x for x in value if x in SAFE_ASCII])

    def getSafe(self, key, default=None, ignore_case=True, data=None, field_type=None):
        value = self.get(key, default, ignore_case, data, field_type)
        if isinstance(value, str):
            return self.removeNonASCII(value)
        return value

    def keys(self):
        if self.__data:
            return list(self.__data.keys())
        return []

    def fromKeys(self, keys):
        # generates a new UberDict, but only with the
        # passed in keys
        d = UberDict()
        for k in keys:
            v = self.get(k, NOTFOUND)
            if v != NOTFOUND:
                d[k] = v
        return d

    def pop(self, key, default=None, field_type=None):
        return self.getStrictType(self.remove(key, default), field_type, default)

    def get(self, key, default=None, ignore_case=True, data=None, field_type=None):
        # lets us pass in a list for keys and get the first match
        if type(key) is list:
            v = None
            for k in key:
                v = self.get(k, NOTFOUND, ignore_case, data, field_type)
                if v != NOTFOUND:
                    return v
            if v == NOTFOUND:
                return default
            return v

        # if type(key) not in [unicode, str]:
        # 	print self.__data
        # 	print "not valid key type: {0} {1}".format(key, type(key))
        # 	return default

        # try json first
        data = self.__data
        if not data:
            # no data in request
            return default

        if "." in key:
            keys = key.split('.')
            for k in keys:
                if ignore_case:
                    data = self.getIgnore(k, data=data)
                else:
                    data = self.getNormal(k, data=data)

                if data is NOTFOUND:
                    return default
            if field_type:
                if data is NOTFOUND or data is None:
                    return default
                return self.getStrictType(data, field_type, default)
            return data

        if ignore_case:
            data = self.getIgnore(key, data=data)
        else:
            data = self.getNormal(key, data=data)

        if data is NOTFOUND:
            return default
        if field_type:
            if data is None:
                return default
            return self.getStrictType(data, field_type, default)
        return data

    def getUserAgent(self):
        return self.request.META.get('HTTP_USER_AGENT', '')

    def getUserAgentPlatform(self):
        ua = self.getUserAgentInfo()
        return f"{ua.user_agent.family} on {ua.os.family}"

    def getUserAgentInfo(self):
        if self.ua_info is None:
            self.ua_info = ua.parse(self.request.META.get('HTTP_USER_AGENT', ''))
        return self.ua_info



from django.contrib import admin
from .models import SMS, PhonenumberInfo
from django import forms


class PhonenumberInfoAdmin(admin.ModelAdmin):
    pass


admin.site.register(PhonenumberInfo, PhonenumberInfoAdmin)
admin.site.register(SMS)




from medialib.models import *
from .render_utils import *
from .exceptions import *
from medialib import stores
from . import schedule
import os
from rest.mail import render_to_mail

from django.conf import settings

from datetime import datetime, timedelta
# import boto
import traceback
import socket
import json
import random

from rest.log import getLogger
logger = getLogger("medialib", filename="medialib.log")


def _log_verbose(msg, always=False):
    logger.info(msg)

def _log_null(msg, always=False):
    # if always:
    # 	print msg
    logger.info(msg)

def do_mail(item, mailtype, message=None):
    if getattr(settings, 'MEDIALIB_NOTIFY_OWNER', False):
        render_to_mail("medialib/mail_" + mailtype, {
            'item': item,
            'message': message,
            'settings': settings,
            'from': settings.MEDIALIB_MAIL_FROM,
            'to': [item.owner.email],
        })

def notify_mail(item, mailtype, message=None):
    render_to_mail("medialib/notify_" + mailtype, {
        'item': item,
        'message': message,
        'settings': settings,
        'from': settings.MEDIALIB_MAIL_FROM,
    })

def start_instance(log=None):
    """
    Start new rendering instance
    """
    if not (hasattr(settings, 'RENDER_AMI') and settings.RENDER_AMI and hasattr(settings, 'RENDER_IMAGE_TYPE') and hasattr(settings, 'RENDER_USER_DATA')):
        if log:
            log("NOT Starting instance")
        return
    # ec2 = boto.connect_ec2(settings.AWS_EC2_KEY, settings.AWS_EC2_SECRET)
    resv = ec2.run_instances(settings.RENDER_AMI, security_groups=settings.RENDER_SECURITY_GROUPS, instance_type=settings.RENDER_IMAGE_TYPE, user_data=settings.RENDER_USER_DATA, instance_initiated_shutdown_behavior='terminate')
    inst = resv.instances[0]
    try:
        inst.add_tag('Name', 'render-starting')
    except:
        pass
    RenderInstance(state="S", instance_id=inst.id).save()
    if log:
        log("Starting instance %s" % inst.id)

def demand_instance(log=None):
    """
    Start new rendering instance if called for
    """
    pending = MediaItem.objects.filter(state=101).count()
    instances = RenderInstance.objects.filter(state__in=('I','S','R',), last_checkin__gt=datetime.now()-timedelta(minutes=10)).count()
    if pending > instances * settings.RENDER_INSTANCE_RATIO:
        start_instance(log)

def rtmpcli_connect(url):
    if not stores.type(url) == "rtmpstore":
        return None

    host = stores.netloc(url)
    if not host:
        return None

    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    try:
        s.connect((host, 1112))
    except OSError:
        return None
    return s

def rtmpcli_do(cmd, url):
    s = rtmpcli_connect(url)
    if not s: return None

    s.send(cmd + "\n")
    s.shutdown(socket.SHUT_WR)
    buf = ""
    while True:
        nbuf = s.recv(1024)
        if nbuf:
            buf += nbuf
        else:
            break
    s.close()
    try:
        ret = json.loads(buf[buf.find('{'):])
    except ValueError:
        return None
    return ret

def live_allow(item):
    if not item.kind == 'L':
        return None

    url = item.live_local_url()
    if not url:
        return None

    ret = rtmpcli_do("start stream=%s" % url.split("/")[-1], url)
    return ret and (ret.get('status', None) == 'SUCCESS')

def live_end(item):
    if not item.kind == 'L':
        return None

    url = item.live_local_url()
    if not url:
        return None

    ret = rtmpcli_do("end stream=%s" % url.split("/")[-1], url)
    return ret and (ret.get('status', None) == 'SUCCESS')

def render_definition(item, rendition, params_override={}, verbose=False, log=None):
    if verbose and not log:
        log = _log_verbose
    if not log:
        log = _log_null

    log("rendering %s for %s" % (rendition.name, rendition.preset.module_name))
    params = {}
    for (n, (p, s, d,),) in list(rendition.getParameters(item=item).items()):
        if s == "" or s is None:
            pass
        elif p.kind == 'I':
            params[n] = int(s)
        elif p.kind == 'F':
            params[n] = float(s)
        elif p.kind == 'B':
            params[n] = bool(s)
        else:
            params[n] = str(s) \
                .replace("$(item)", str(item.pk)) \
                .replace("$(rendition)", str(rendition.pk)) \
                .replace("$(random)", str(random.randint(10000000, 99999999)))
    # print params_override
    # print params
    for p in params_override:
        params[p] = params_override[p]

    logger.info("params", params)
    try:
        back = __import__('medialib.render.presets.{0}'.format(rendition.preset.module_name), globals(), locals(), ['render'])
    except ImportError:
        logger.exception("Missing rendition back end for '%s'" % rendition.preset.module_name)
        raise RenderError("Missing rendition back end for %s" % rendition.preset.module_name)
    try:
        back.render(item, rendition, params)
        logger.info("rendering returned with no errors")
    except Exception:
        logger.exception("render_definition")

def render(item, stage=None, files={}, verbose=False, log=None, definition=None, use=None, params_override={}):
    """
    | Parameter: stage: rendition stage number (1 = initial, >1 = subsequent)
    | Parameter: files: manually specify pre-downloaded file
    | Parameter: verbose: verbose output
    | Parameter: log: logging function
    | Parameter: definition: render definition with specified name (in any rendition set)
    | Parameter: use: render presets with specified use (in handling rendition set)
    | Parameter: params_override: dictionary of override parameter settings

    | Return: REST list of all MediaLibrary data

    | Render media item
    """
    if verbose and not log:
        log = _log_verbose
    if not log:
        log = _log_null
    from ..models import MediaItemRendition, RenditionSet, MediaItem, RenditionDefinition
    ok = True
    try:
        try:
            # 10 minutes to upload original
            if not item.original().check_file():
                if (datetime.now() - item.created).total_seconds() > 600:
                    raise RenderError("Missing original rendition in storage")
                log("Skipped: original rendition not ready")
                return False
        except MediaItemRendition.DoesNotExist:
            if (datetime.now() - item.created).total_seconds() > 600:
                raise RenderError("Missing original rendition in DB")
            log("Skipped: original rendition not ready")
            return False

        if item.state == 100:
            isnew = True
            live_allow(item)
        else:
            isnew = False

        if item.state >= 100 and item.state <= 120:
            item.saveState(102)

        if definition:
            renditions = RenditionDefinition.objects.filter(active=True, name=definition)
            rset = None
        else:

            rset = item.library.rendition_sets.filter(kind=item.kind).first()
            if rset is None:
                rset = RenditionSet.objects.filter(kind=item.kind, default_set=True).first()
                if rset is None:
                    log("Invalid file kind: {0}".format(item.kind))
                    raise ValidateError("Invalid file type")
            renditions = rset.renditions.filter(active=True)
            if use:
                renditions = renditions.filter(use=use)
            # if stage:
            #     renditions = renditions.filter(preset__stage__gte=((stage-1)*100), preset__stage__lt=(stage*100))

        renditions = renditions.order_by('preset__stage', 'id')

        for f in files:
            get_rendition(item, f, fp=files[f])

        for rendition in renditions:
            render_definition(item, rendition, params_override=params_override, log=log)
        schedule.run()
        MediaItem.objects.update() # invalidate cache
        item = MediaItem.objects.get(pk=item.pk)
        if item.state >= 100 and item.state <= 120:
            if stage and rset and rset.renditions.filter(active=True).filter(preset__stage__gte=((stage)*100)).count() > 0:
                item.saveState(101)
            elif item.state != 200:
                item.render_error = None
                item.saveState(200)
                if not isnew and item.kind != 'L':
                    do_mail(item, 'ready')
                    if getattr(settings, 'MEDIALIB_NOTIFY_READY', False):
                        notify_mail(item, 'ready')
        if item.state >= 100 and item.state <= 120:
            if isnew:
                do_mail(item, 'pending')
                if getattr(settings, 'MEDIALIB_NOTIFY_PENDING', False):
                    notify_mail(item, 'pending')
            # demand_instance(log)
            # log("demanding instance but not allowed: {} {}".format(item.kind, item.state))
        if item.kind == 'L' and item.state >= 110:
            live_end(item)
            if item.state == 200:
                if isnew:
                    item.state = 100
                else:
                    item.state = 101
                item.kind = 'A'
                item.saveState(item.state)
                return render(item, 1, files=files, verbose=verbose, log=log, definition=definition, use=use, params_override=params_override)

    except ValidateError as err:
        item.render_error = err.message
        log(item.render_error, True)
        item.state = 111
        item.saveState(item.state)
        notify_mail(item, 'validate_error', item.render_error)
        return False
    except RenderError as err:
        item.render_error = err.message
        log(item.render_error, True)
        if not definition:
            item.state = 110
        item.saveState(item.state)
        do_mail(item, 'error', item.render_error)
        notify_mail(item, 'render_error', item.render_error)
        return False
    except:
        item.render_error = traceback.format_exc()
        log(item.render_error, True)
        if not definition:
            item.state = 110
        item.saveState(item.state)
        do_mail(item, 'error', item.render_error)
        notify_mail(item, 'error', item.render_error)
        return False
    finally:
        close_files(item)

    return True





import requests

def getRandomJoke():
    r = requests.get('https://icanhazdadjoke.com/', headers={"Accept":"application/json"})
    try:
        data = r.json()
        return data.get("joke", "failed")
    except:
        pass
    return "Tired, try later."



from django.db import models
from rest import settings

from rest import models as rm
from rest import helpers as rh
from taskqueue.models import Task
from account.models import Member
from objict import objict
from datetime import datetime, timedelta
from rest import log
from ws4redis import client as ws4redis
import time

logger = log.getLogger("incident", filename="incident.log")

EVENT_META_TO_INCIDENT = [
    "username",
    "email",
    "phone",
    "group_name",
    "group_id",
    "http_user_agent",
    "path"
]

INCIDENT_STATE_NEW = 0
INCIDENT_STATE_OPENED = 1
INCIDENT_STATE_PAUSED = 2
INCIDENT_STATE_IGNORE = 3
INCIDENT_STATE_RESOLVED = 4
INCIDENT_STATE_PENDING = 5

INCIDENT_STATES = [
    (INCIDENT_STATE_NEW, "new"),
    (INCIDENT_STATE_OPENED, "opened"),
    (INCIDENT_STATE_PAUSED, "paused"),
    (INCIDENT_STATE_IGNORE, "ignored"),
    (INCIDENT_STATE_RESOLVED, "resolved"),
    (INCIDENT_STATE_PENDING, "pending"),
]

INCIDENT_EMAIL_FROM = settings.get("INCIDENT_EMAIL_FROM", None)
if INCIDENT_EMAIL_FROM is None:
    INCIDENT_EMAIL_FROM = f"{settings.SITE_LABEL} INCIDENT <incident@{settings.EMAIL_DEFAULT_DOMAIN}>"


class Incident(models.Model, rm.RestModel, rm.MetaDataModel):
    class RestMeta:
        POST_SAVE_FIELDS = ["level", "catagory", "action"]
        SEARCH_FIELDS = ["description", "hostname"]
        VIEW_PERMS = ["view_incidents", "view_issues"]
        LIST_PARENT_KINDS = ["org", "iso"]
        LIST_CHILD_DEPTH = 3
        GRAPHS = {
            "default": {
                "extra": ["metadata", ("get_state_display", "state_display")],
                "graphs": {
                    "group": "basic",
                    "rule": "basic",
                    "assigned_to": "basic"
                },
            }
        }

    created = models.DateTimeField(auto_now_add=True)
    modified = models.DateTimeField(auto_now=True)

    description = models.CharField(max_length=200)

    category = models.CharField(max_length=124, null=True, default=None, db_index=True)

    component = models.CharField(max_length=200, null=True, default=None, db_index=True)
    component_id = models.IntegerField(null=True, blank=True, default=None, db_index=True)

    hostname = models.CharField(max_length=200, null=True, default=None, db_index=True)
    reporter_ip = models.CharField(max_length=16, blank=True, null=True, default=None, db_index=True)

    group = models.ForeignKey("account.Group", on_delete=models.SET_NULL, null=True, default=None)
    assigned_to = models.ForeignKey("account.Member", on_delete=models.SET_NULL, null=True, default=None)
    
    priority = models.IntegerField(default=0)  # 1-10, 1 being the highest
    # 0=new, 1=opened, 2=paused, 3=ignore, 4=resolved, 5=pending
    state = models.IntegerField(default=0, choices=INCIDENT_STATES)
    action_sent = models.DateTimeField(default=None, null=True)

    rule = models.ForeignKey("incident.Rule", on_delete=models.SET_NULL, null=True, default=None)

    @property
    def first_event(self):
        return self.events.first()

    @property
    def action_triggered(self):
        if self.rule is None or self.rule.action_after == 0:
            return True
        return self.events.all().count() >= abs(self.rule.action_after)

    @property
    def bundle_expired(self):
        if self.rule is not None:
            return self.created < (datetime.now() - timedelta(minutes=self.rule.bundle))
        return True

    def set_action(self, value):
        if value == "merge":
            request = self.getActiveRequest()
            ids = request.DATA.getList("merge_ids")
            if ids and len(ids) > 0:
                self.mergeWith(ids, request)
                self._changed__ = objict()

    def mergeWith(self, ids, request=None):
        qset = Incident.objects.filter(pk__in=ids).exclude(pk=self.pk)
        for incident in qset:
            self.logHistory(note=f"merged Incident #{incident.pk}", request=request, member=request.member)
            for event in incident.events.all():
                event.incident = self
                event.save()
            for h in self.history.all():
                h.parent = self
                h.save()
        qset.delete()
        count = self.events.all().count()
        self.setProperty("event_count", count)

    def updateMeta(self, event):
        for key in EVENT_META_TO_INCIDENT:
            if key in event.metadata:
                self.setProperty(key, event.metadata[key])
        self.setProperty("details", event.details)
        if event.metadata.geoip and event.metadata.geoip.city:
            self.setProperty("city", event.metadata.geoip.city)
            self.setProperty("state", event.metadata.geoip.state)
            self.setProperty("country", event.metadata.geoip.country)

    def shouldTriggerAction(self):
        count = self.events.all().count()
        self.setProperty("event_count", count)
        aa = self.rule.action_after
        if aa >= 0:
            return aa == count-1
        aa = abs(self.rule.action_after)
        return (count % aa) == 0

    def triggerAction(self, force=False):
        if self.rule is None:
            if self.action_sent is None:
                self.triggerAsyncNotify()
                self.triggerNotify()
            return
        # if self.rule.bundle == 0:
        #     return
        # only do query if not 0
        logger.info("triggerAction", self.rule.action)
        if force or self.shouldTriggerAction():
            logger.info(f"triggering incident action: {self.rule.action}")
            if self.state == INCIDENT_STATE_PENDING:
                self.state = INCIDENT_STATE_NEW
                self.save()
            self.triggerAsyncNotify()
            if self.rule.action is None or self.rule.action == "notify":
                self.triggerNotify()
            elif self.rule.action.startswith("email:") or self.rule.action.startswith("notify:"):
                self.triggerEmail()
            elif self.rule.action.startswith("sms:"):
                self.triggerSMS()
            elif self.rule.action.startswith("task:"):
                self.triggerTask()
            elif self.rule.action.startswith("group:"):
                self.triggerGroup()
            elif self.rule.action == "resolved":
                self.state = INCIDENT_STATE_RESOLVED
                self.save()
            elif self.rule.action.startswith("firewall_block"):
                if settings.FIREWALL_GLOBAL_BLOCK:
                    Task.Publish("incident", "firewall_block", 
                                 dict(ip=self.reporter_ip),
                                 channel="tq_broadcast")

    def triggerAsyncNotify(self):
        if self.lastSentAge() < 300:
            return
        msg = dict(
            pk=self.pk,
            created=time.mktime(self.created.timetuple()),
            description=self.description,
            category=self.category,
            component=self.component,
            component_id=self.component_id,
            hostname=self.hostname)
        if self.rule is not None:
            msg["rule"] = self.rule.name
        event = self.first_event
        if event is not None:
            msg["event_count"] = self.getProperty("event_count", default=1)
            msg["catagory"] = event.category
            msg["details"] = event.getProperty("details")
            msg["hostname"] = event.hostname
            msg["username"] = event.getProperty("username")
            msg["server"] = event.getProperty("server")
            msg["ip"] = event.getProperty("ip")
            msg["method"] = event.getProperty("method")
        if self.group is not None:
            msg["group_id"] = self.group.id
        try:
            ws4redis.sendMessageToPK("incident", "all", msg)
        except Exception:
            rh.log_exception("triggerAsyncNotify")

    def triggerTask(self):
        # task:APP_NAME:FNAME:CHANNEL
        fields = self.rule.action.split(':')
        if len(fields) < 4:
            rh.log_error("triggerTask failed, invalid field count")
            return
        self.action_sent = datetime.now()
        self.save()
        Task.Publish(fields[1], fields[2], channel=fields[3])

    def triggerEmail(self):
        if self.lastSentAge() < 300:
            return
        # email:NOTIFY_SETTING or email:bob@example.com
        action, perm = self.rule.action.split(":")
        self.action_sent = datetime.now()
        self.save()
        # notify with perm
        self.notifyWith(perm)

    def triggerSMS(self):
        if self.lastSentAge() < 300:
            return
        # sms:NOTIFY_SETTING
        self.action_sent = datetime.now()
        self.save()
        try:
            action, perm = self.rule.action.split(":")
            members = Member.GetWithNotification(perm)
            # count = self.getProperty("event_count", default=1, field_type=int)
            url = F"{settings.INCIDENT_PORTAL_URL}?incident={self.pk}"
            msg = f"Incident #{self.pk}\n{self.description}\n{url}"
            for m in members:
                m.sendSMS(msg)
        except Exception:
            rh.log_exception("triggerSMS")

    def triggerGroup(self):
        if not self.group:
            self.notifyWith("notify.unknown_incidents")
            return
        action, perm = self.rule.action.split(":")
        self.action_sent = datetime.now()
        self.save()

        subject = f"New Issue @ {self.group.name} - {self.description}"
        details = self.getProperty("details", self.description)
        username = self.getProperty("username", None)
        email = self.getProperty("email", None)
        phone = self.getProperty("phone", None)
        if self.category == "support":
            sms_msg = f"{settings.SITE_LABEL}\n{self.description}\n{details}"
        else:
            sms_msg = f"{settings.SITE_LABEL}\nIssue @ {self.group.name}\n{details}"

        body = []

        if username is not None:
            body.append(f"Reported by: {username}")
        if phone is not None:
            body.append(f"Requesting Callback: {phone}")
        elif email is not None:
            body.append(f"Respond to: {email}")
        body.append(f"Location: {self.group.name}")
        body.append("<hr>")
        body.append(f"<pre>{details}</pre>")
        if username is not None:
            body.append(f"- {username}")
        body = "<br>\n".join(body)
        # now we loop up through our groups parents
        count = 0
        group = self.group
        if perm.startswith("notify."):
            perm = perm.split(".")[1]
        sms_perms = f"notify.sms_{perm}"
        email_perms = f"notify.email_{perm}"
        while group is not None or count > 4:
            self.notifyMembers(group, subject, body, sms_msg, sms_perms, email_perms)
            group = group.parent

    def notifyMembers(self, group, subject, body, sms_msg, sms_perms, email_perms):
        members = group.getMembers(perms=sms_perms, as_member=True)
        for member in members:
            member.sendSMS(sms_msg)
        members = group.getMembers(perms=email_perms, as_member=True)
        for member in members:
            member.sendEmail(subject, body)       

    def notifyWith(self, perm):
        # logger.info("notifyWith", perm)
        count = self.getProperty("event_count", default=1, field_type=int)
        events = None
        if count > 1:
            subject = F"Incident #{self.pk} {self.category}@{self.hostname} had {count} events"
            events = self.events.all()
        else:
            subject = F"New Incident #{self.pk} {self.category}@{self.hostname}"

        Member.notifyWith(
            perm,
            subject=subject,
            template=settings.get("INCIDENT_TEMPLATE", "email/incident_plain.html"),
            context=dict(incident=self, portal_url=settings.INCIDENT_PORTAL_URL),
            email_only=True, from_email=INCIDENT_EMAIL_FROM)

    def lastSentAge(self):
        # prevent spams, only allow emails every 5 minutes
        if self.action_sent is None:
            return 100000
        return int((datetime.now() - self.action_sent).total_seconds())

    def triggerNotify(self):
        if self.lastSentAge() < 300:
            return
        self.action_sent = datetime.now()
        self.save()
        self.notifyWith("notify.unknown_incidents")

    def on_rest_saved(self, request, is_new=False):
        if not is_new:
            if self._changed__:
                self.logHistory(request=request)
            if request != None and len(request.FILES):
                for name, value in request.FILES.items():
                    self.logHistory(kind="media", media=value, request=request)
            if request != None and "DATA" in request and "note" in request.DATA:
                self.logHistory(kind="note", note=request.DATA.get("note"), request=request)

    def logHistory(self, kind="history", note=None, media=None, 
                   request=None, member=None, notify=True):
        if request is None:
            request = self.getActiveRequest()
        if member is None and request is not None and hasattr(request, "member"):
            member = request.member
        if note is None and self.has_model_changed:
            notes = []
            for k, v in self._changed__.items():
                nv = self.getFieldValue(k)
                notes.append(f"{k} changed from {v} to {nv}")
            note = "\n<br>".join(notes)

        h = IncidentHistory(
            parent=self,
            to=self.assigned_to,
            note=note,
            kind=kind,
            priority=self.priority,
            state=self.state)
        if member is not None:
            h.by = member
            if self.assigned_to is None:
                self.assigned_to = member
                self.save()
        if media is not None:
            h.saveMediaFile(media, "media", media.name)
        h.save()
        if notify and h.state != INCIDENT_STATE_IGNORE:
            self.notifyWatchers(
                subject=F"Updated Incident #{self.id}",
                history=h)

    def notifyWatchers(self, subject, history=None):
        action = None
        perm = "notify.incident_alerts"
        if self.rule is not None and (self.rule.action.startswith("email:") or self.rule.action.startswith("notify:")):
            action, perm = self.rule.action.split(":")
        context = dict(
                    incident=self,
                    portal_url=settings.INCIDENT_PORTAL_URL,
                    history=history)
        template = "email/incident_plain.html"

        # this should notify all users in our incident group of the change
        if self.group is not None:
            # all member of the group are notified because it is an incident group
            self.group.notifyMembers(
                subject=subject,
                template=template,
                context=context,
                perms=[perm],
                email_only=True,
                from_email=INCIDENT_EMAIL_FROM)
        elif history.by is None:
            # notify everyone with the perm
            Member.notifyWith(
                perm,
                subject,
                template=template,
                context=context,
                email_only=True,
                from_email=INCIDENT_EMAIL_FROM)
        else:
            # notitfy everyone but the sender          
            if history.by is None:
                members = Member.GetWithPermission(perm).exclude(pk=history.by.pk)
                if members.count() == 0:
                    return
                NotificationRecord = Incident.getModel("account", "NotificationRecord")
                NotificationRecord.notify(
                    members,
                    subject,
                    template=template,
                    context=context,
                    email_only=True,
                    from_email=INCIDENT_EMAIL_FROM)

    @classmethod
    def getBundled(cls, rule, event):
        # calculate our bundle start time
        when = datetime.now() - timedelta(minutes=rule.bundle)
        q = objict(rule=rule, created__gte=when)
        if rule.bundle_by in [2, 3, 5]:
            if event.component_id:
                q.component = event.component
                q.component_id = event.component_id
        if rule.bundle_by in [1, 3, 7]:
            q.hostname = event.hostname
        if rule.bundle_by in [4, 5, 8]:
            q.reporter_ip = event.reporter_ip
        if rule.bundle_by in [6, 7, 8]:
            q.category = event.category
        return Incident.objects.filter(**q).last()

    @classmethod
    def canPublishTo(cls, credentials, msg):
        if credentials:
            return True
        return False


class IncidentMetaData(rm.MetaDataBase):
    parent = models.ForeignKey(Incident, related_name="properties", on_delete=models.CASCADE)


class IncidentHistory(models.Model, rm.RestModel):
    class Meta:
        ordering = ['-created']

    class RestMeta:
        SEARCH_FIELDS = ["to__username", "note"]
        GRAPHS = {
            "default": {
                "extra":[
                    ("get_state_display", "state_display"),
                    ("get_priority_display", "priority_display"),
                ],
                "graphs":{
                    "by":"basic",
                    "to":"basic",
                    "media": "basic"
                }
            },
        }
    parent = models.ForeignKey(Incident, related_name="history", on_delete=models.CASCADE)
    created = models.DateTimeField(auto_now_add=True, editable=False)

    group = models.ForeignKey("account.Group", blank=True, null=True, default=None, related_name="+", on_delete=models.CASCADE)

    kind = models.CharField(max_length=80, blank=True, null=True, default=None, db_index=True)

    to = models.ForeignKey("account.Member", blank=True, null=True, default=None, related_name="+", on_delete=models.CASCADE)
    by = models.ForeignKey("account.Member", blank=True, null=True, default=None, related_name="+", on_delete=models.CASCADE)

    state = models.IntegerField(default=0)
    priority = models.IntegerField(default=0)

    note = models.TextField(blank=True, null=True, default=None)
    media = models.ForeignKey("medialib.MediaItem", related_name="+", null=True, default=None, on_delete=models.CASCADE)




# coding: utf-8

"""
    mistune_contrib.math
    ~~~~~~~~~~~~~~~~~~~~

    Support Math features for mistune.

    :copyright: (c) 2014 by Hsiaoming Yang.
"""

import re


class MathBlockMixin(object):
    """Math mixin for BlockLexer, mix this with BlockLexer::

        class MathBlockLexer(MathBlockMixin, BlockLexer):
            def __init__(self, *args, **kwargs):
                self.enable_math()
                super(MathBlockLexer, self).__init__(*args, **kwargs)
    """
    def enable_math(self):
        self.rules.block_math = re.compile(r'^\$\$(.*?)\$\$', re.DOTALL)
        self.rules.block_latex = re.compile(
            r'^\\begin\{([a-z]*\*?)\}(.*?)\\end\{\1\}', re.DOTALL
        )
        self.default_rules.extend(['block_math', 'block_latex'])

    def parse_block_math(self, m):
        """Parse a $$math$$ block"""
        self.tokens.append({
            'type': 'block_math',
            'text': m.group(1)
        })

    def parse_block_latex(self, m):
        self.tokens.append({
            'type': 'block_latex',
            'name': m.group(1),
            'text': m.group(2)
        })


class MathInlineMixin(object):
    """Math mixin for InlineLexer, mix this with InlineLexer::

        class MathInlineLexer(InlineLexer, MathInlineMixin):
            def __init__(self, *args, **kwargs):
                self.enable_math()
                super(MathInlineLexer, self).__init__(*args, **kwargs)
    """

    def enable_math(self):
        self.rules.math = re.compile(r'^\$(.+?)\$')
        self.default_rules.insert(0, 'math')

    def output_math(self, m):
        return self.renderer.math(m.group(1))


class MathRendererMixin(object):
    def block_math(self, text):
        return '$$%s$$' % text

    def block_latex(self, name, text):
        return r'\begin{%s}%s\end{%s}' % (name, text, name)

    def math(self, text):
        return '$%s$' % text



from rest import decorators as rd
from rest import views as rv
from wiki import models as wiki


@rd.url('page')
@rd.url('page/<int:pk>')
# @rd.login_required
def rest_on_manage_wiki(request, pk=None):
    return wiki.Page.on_rest_request(request, pk)


@rd.url('path')
@rd.url('path/<path:path>')
# @rd.login_required
def rest_on_wiki(request, path=None):
    if path:
        entry = wiki.Page.objects.filter(path=path).last()
        if entry is None:
            # try to find parent
            if request.method == "POST":
                parent = wiki.Page.path_parent(path)
                if parent is not None:
                    slug = path.split("/").pop()
                    entry = wiki.Page.createFromRequest(request, parent=parent, slug=slug)
                    return entry.on_rest_get(request) 
            return rv.restNotFound(request)
        if request.method == "GET":
            return entry.on_rest_get(request)
        elif request.method == "POST":
            return entry.on_rest_post(request)
        elif request.method == "DELETE":
            return entry.on_rest_delete(request)
    return wiki.Page.on_rest_request(request, None)


@rd.url('media')
@rd.url('media/<int:pk>')
# @rd.login_required
def rest_on_wiki_media(request, pk=None):
    return wiki.PageMedia.on_rest_request(request, pk)



# Generated by Django 4.1.4 on 2023-11-15 20:20

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('account', '0015_memberdevice_buid'),
    ]

    operations = [
        migrations.AddField(
            model_name='authsession',
            name='buid',
            field=models.CharField(blank=True, default=None, max_length=127, null=True),
        ),
    ]



from django.db import models
from django.db.models import Q
import re

from rest.models import RestModel, MetaDataModel, MetaDataBase, RestValidationError, PermissionDeniedException
from rest import helpers as rh
from rest import RemoteEvents
from rest import crypto
from rest import mail as rest_mail
from rest.views import restPermissionDenied

from .member import Member

from rest import settings
app_settings = settings.getAppSettings("account")
MEMBERSHIP_ROLES = settings.get("MEMBERSHIP_ROLES", None)
TRUE_VALUES = app_settings.get("TRUE_VALUES", [])


class Group(models.Model, RestModel, MetaDataModel):
    """
    Group Model allows for the grouping of other models and works with Member throug Membership Model

    parent allows for tree based heirachy of groups
    children allows for manytomany relationships with other groups
    kind is heavily used to filter different kinds of groups
    """
    created = models.DateTimeField(auto_now_add=True)
    modified = models.DateTimeField(auto_now=True)

    uuid = models.CharField(db_index=True, max_length=64, blank=True, null=True, default=None)
    name = models.CharField(db_index=True, max_length=200)
    short_name = models.CharField(max_length=60, null=True, blank=True, default=None)
    kind = models.CharField(db_index=True, max_length=80, default="org")
    parent = models.ForeignKey("Group", default=None, null=True, blank=True, related_name="groups", on_delete=models.CASCADE)

    is_active = models.BooleanField(default=True, blank=True)
    # this is the new model for groups having multiple parents
    children = models.ManyToManyField("self", related_name="parents", symmetrical=False)

    location = models.ForeignKey(
        "location.Address", default=None, null=True, blank=True,
        related_name="+", on_delete=models.SET_NULL)

    class RestMeta:
        SEARCH_FIELDS = [
            "name"
        ]
        LIST_DEFAULT_FILTERS = {
            "is_active": True
        }
        LIST_PARENT_KINDS = ["org", "iso"]
        POST_SAVE_FIELDS = ["child_of"]
        GROUP_FIELD = "self"
        VIEW_PERMS = [
            "view_all_groups",
            "manage_groups",
            "manage_group",
            "manage_settings"]
        SAVE_PERMS = [
            "manage_groups",
            "manage_group",
            "manage_settings"]
        CREATE_PERMS = ["manage_groups", "create_groups"]
        GRAPHS = {
            "basic": {
                "fields": [
                    "id",
                    "parent",
                    "uuid",
                    "name",
                    "short_name",
                    "kind",
                    "created",
                    "thumbnail",
                    "is_active",
                    "timezone"
                ]
            },
            "default": {
                "graphs": {
                    "self": "basic",
                    "parent": "basic",
                    "location": "basic"
                },
                "fields": ["metadata"],
            },
            "detailed": {
                "graphs": {
                    "self": "basic",
                    "parent": "basic",
                    "children": "basic",
                    "location": "basic"
                },
                "extra": ["metadata"],
            },
            "location": {
                "fields": ["name", "kind", "id"],
                "extra": ["location"],
                "graphs": {
                    "location": "basic"
                }
            },
            "abstract": {
                "fields": [
                    ('uuid', 'id'),
                    "name",
                    "kind",
                    "timezone"
                ]
            }
        }
        FORMATS = {
            "csv": [
                "created",
                "id",
                "name",
                "kind",
                ("metadata__timezone", "timezone"),
                ("metadata__eod", "end_of_day"),
                ("parent__name", "parent"),
                ("metadata.location.line1", "address_line1"),
                ("metadata.location.line2", "address_line2"),
                ("metadata.location.city", "city"),
                ("metadata.location.state", "state"),
                ("metadata.location.country", "country"),
                ("metadata.location.zip", "zipcode")
            ]
        }

    @property
    def timezone(self):
        return self.getProperty("timezone", "America/Los_Angeles")

    @property
    def timezone_short(self):
        zone = self.getProperty("timezone", "America/Los_Angeles")
        return rh.getShortTZ(zone)

    @property
    def file_safe_name(self):
        return re.sub("[^0-9a-zA-Z]+", "_", self.name.lower())

    @property
    def initials(self):
        if self.name:
            return "".join([s[0] for s in self.name.split(' ')]).upper()
        return None

    def thumbnail(self, name="default"):
        lib = self.libraries.filter(name=name).first()
        if lib:
            item = lib.items.all().first()
            if item:
                return item.thumbnail_url()
        return None
        
    @classmethod
    def on_rest_list_filter(cls, request, qset=None):
        if not request.member.hasPermission("view_all_groups"):
            qset = request.member.getGroups()
        # override on do any pre filters
        child_of = request.DATA.get("child_of")
        if request.group is not None and child_of is None:
            child_of = request.group.id
        parent_id = request.DATA.get(["parent", "parent_id"])
        if parent_id:
            parent = request.member.getGroup(parent_id)
            if not parent:
                raise PermissionDeniedException("invalid parent")
            qset = qset.filter(parent=parent)
        elif child_of:
            parent = request.member.getGroup(child_of)
            if parent:
                request.group = None
                return parent.getAllChildren(grand_children=True)
        return qset

    def on_rest_get(self, request):
        if (request.member and request.member.isMemberOf(self)) or self.on_rest_can_get(request):
            return self.restGet(request)
        return restPermissionDenied(request)

    # def onRestCanSave(self, request=None, data=None, extended=None):
    #     if request.member is None:
    #         raise PermissionDeniedException("permission denied for save")
    #     if request.member.hasPermission(["manage_groups", "create_groups"]):
    #         return True
    #     if self.checkPermission(request.member, ["manage_settings", "manage_members", "manage_group"]):
    #         return True
    #     raise PermissionDeniedException("permission denied for save")

    def on_rest_pre_save(self, request, **kwargs):
        pass

    def on_rest_saved(self, request, is_new=False):
        if request.member:
            note = "edited group {}:{}\n{}".format(self.name, self.pk, request.DATA.asDict())
            request.member.log("group_edit", note, method="group")
            self.auditLog(note, "group_edit", level=21)
            self.logEvent("group_edit", component="account.Member", component_id=request.member.id, note=note)

    def set_child_of(self, value):
        # this is a helper to add this group to another group
        parent = Group.objects.filter(pk=value).last()
        if parent and parent.pk != self.pk:
            if not parent.children.filter(pk=self.pk).exists() and not self.hasChild(parent) and self.kind != "org":
                parent.children.add(self)

    def set_remove_parent(self, value):
        parent = Group.objects.filter(pk=value).last()
        if parent:
            if parent.children.filter(pk=self.pk).exists():
                parent.children.remove(self)

    def set_timezone(self, value):
        if self.pk:
            self.setProperty("timezone", value)

    def getChildModels(self, Model, grand_children=False, great_grand_children=False):
        q = Q(group=self)
        if grand_children:
            q = q | Q(group__parent=self)
        if great_grand_children:
            q = q | Q(group__parent__parent=self)
        if hasattr(Model, "objects"):
            return Model.objects.filter(q)
        return Model.filter(q)

    def getChildren(self):
        return Group.objects.filter(parent=self)

    def getAllChildren(self, include_me=False, grand_children=False, great_grand_children=False):
        q = Q(parent=self)| Q(parents=self)
        if include_me:
            q = q | Q(pk=self.id)
        if grand_children:
            q = q | Q(parent__parent=self)
        if great_grand_children:
            q = q | Q(parent__parent__parent=self)
        return Group.objects.filter(q)

    def getAllChildrenIds(self, include_me=False, grand_children=False, great_grand_children=False, depth=0):
        if depth > 1:
            grand_children = True
        if depth > 2:
            great_grand_children = True
        qset = self.getAllChildren(include_me, grand_children, great_grand_children)
        return list(qset.values_list("id", flat=True))

    def hasChild(self, group):
        if not group:
            return False
        if self.children.filter(pk=group.pk).exists():
            return True
        for child in self.children.all():
            if child.hasChild(group):
                return True
        return False

    def getParentOfKind(self, kind):
        if self.parent and self.parent.kind == kind:
            return self.parent
        group = self.parents.filter(kind=kind).first()
        if group:
            return group
        for parent in self.parents.all():
            if parent.kind == kind:
                return parent
            group = parent.getParentOfKind(kind)
            if group:
                return group
        return None

    def hasParent(self, group):
        # this needs to check parents...then check each parent for parent
        if self.parent == group:
            return True
        if self.parents.filter(pk=group.id).count():
            return True
        for parent in self.parents.all():
            if parent == group:
                return True
            if parent.hasParent(group):
                return True
        return False

    def notifyMembers(self, subject, message=None, template=None, context=None, 
                      email_only=False, sms_msg=None, perms=None, force=False,
                      from_email=None):
        if perms is not None:
            members = self.getMembers(perms=perms, as_member=True)
        else:
            Member = RestModel.getModel("account", "Member")
            members = Member.objects.filter(is_active=True, memberships__group=self, memberships__state__gte=-10)
        NotificationRecord = RestModel.getModel("account", "NotificationRecord")
        NotificationRecord.notify(
            members, subject, message, template,
            context, email_only, sms_msg, force,
            from_email=from_email)

    def hasPerm(self, member, perm, staff_override=True, check_member=False):
        return self.checkPermission(member, perm, staff_override, check_member)

    def checkPermission(self, member, perm, staff_override=True, check_member=False):
        if member.is_superuser:
            return True
        if staff_override and member.is_staff:
            return True
        if check_member:
            if member.hasPerm(perm) or member.hasGroupPerm(self, perm):
                return True
        memberships = member.memberships.filter(group=self)
        for ms in memberships:
            if ms.hasPermission(perm):
                return True
        return False

    def getLocalTime(self, when=None, tz_aware=False):
        zone = self.getProperty("timezone", "America/Los_Angeles")
        return rh.convertToLocalTime(zone, when, tz_aware)

    def getUTC(self, when):
        zone = self.getProperty("timezone", "America/Los_Angeles")
        return rh.convertToUTC(zone, when)

    def getBusinessDay(self, start=None, end=None, kind="day"):
        zone = self.getProperty("timezone", "America/Los_Angeles")
        eod = self.getProperty("eod", 0, field_type=int)
        return rh.getDateRange(start, end, kind, zone, hour=eod)

    def getOperatingHours(self, start=None, end=None, kind="day"):
        # deprecate this, operating hours is deceptive
        zone = self.getProperty("timezone", "America/Los_Angeles")
        eod = self.getProperty("eod", 0, field_type=int)
        return rh.getDateRange(start, end, kind, zone, hour=eod)

    def getTimeZoneOffset(self, when=None, hour=None):
        zone = self.getProperty("timezone", "America/Los_Angeles")
        return rh.getTimeZoneOffset(zone, when, hour=hour)

    def getEOD(self, eod=None, onday=None, in_local=False):
        if eod is None:
            eod = self.getProperty("eod", 0, field_type=int)
            if in_local:
                return eod
        offset = self.getTimeZoneOffset(onday, hour=eod)
        return offset

    def updateUUID(self):
        self.uuid = crypto.obfuscateID("group", self.id)
        self.save()

    def isMember(self, member):
        return self.memberships.filter(member=member, state__gte=-10).count()

    def hasMember(self, member):
        return self.isMember(member)

    def addMember(self, member, role):
        return self.addMembership(member, role)

    def addMembership(self, member, role):
        if self.memberships.filter(member=member, role=role).count():
            return None
        Membership = RestModel.getModel("account", "Membership")
        ms = Membership(group=self, member=member, role=role)
        ms.save()
        return ms

    def getMembers(self, perms=None, role=None, as_member=False):
        if isinstance(perms, str):
            perms = [perms]
        if isinstance(role, str):
            role = [role]
        notify = None
        if perms:
            notify = [p.split('.')[1] for p in perms if p.startswith("notify.")]
            perms = [p for p in perms if not p.startswith("notify.")]
        if as_member:
            Member = RestModel.getModel("account", "Member")
            res = Member.objects.filter(is_active=True, memberships__group=self, memberships__state__gte=-10)
            if perms:
                res = res.filter(memberships__group=self, memberships__properties__category="permissions", memberships__properties__key__in=perms, memberships__properties__value__in=TRUE_VALUES)
            if notify:
                res = res.filter(memberships__group=self, memberships__properties__category="notify", memberships__properties__key__in=notify, memberships__properties__value__in=TRUE_VALUES)
            if role:
                res = res.filter(memberships__group=self, memberships__role__in=role)
            return res.distinct()
        res = self.memberships.filter(state__gte=-10)
        if perms:
            res = res.filter(
                Q(properties__category="permissions", 
                  properties__key__in=perms, 
                  properties__value__in=TRUE_VALUES) | 
                Q(member__properties__category="permissions", 
                  member__properties__key__in=perms, 
                  member__properties__value__in=TRUE_VALUES))
        if notify:
            res = res.filter(
                Q(properties__category="notify", 
                  properties__key__in=notify, 
                  properties__value__in=TRUE_VALUES) | 
                Q(member__properties__category="notify", 
                  member__properties__key__in=notify, 
                  member__properties__value__in=TRUE_VALUES))

        if role:
            res = res.filter(role__in=role)
        return res.distinct()

    def getParentIDs(self):
        ids = []
        group = self
        counter = 0
        while group.parent != None and counter < 6:
            counter += 1
            ids.append(group.parent.id)
            group = group.parent
        return ids

    def getMembership(self, member, include_parents=False):
        ms = self.memberships.filter(member=member).first()
        if ms or not include_parents:
            return ms
        group = self
        counter = 0
        while group.parent != None and counter < 6:
            counter += 1
            group = group.parent
            ms = group.getMembership(member)
            if ms:
                return ms
        return ms

    def invite(self, member, role="guest"):
        # invite a user to this group
        Membership = RestModel.getModel("account", "Membership")
        ms = Membership.objects.filter(group=self, member=member).last()
        if ms is None:
            ms = Membership(member=member, group=self, role=role)
            ms.save()
        elif ms.role != role:
            ms.clearPermissions()
        if not ms.is_enabled:
            # enable again?
            ms.set_state(0)
            ms.save()
            
        if MEMBERSHIP_ROLES:
            perms = MEMBERSHIP_ROLES.get(role, [])
            for k in perms:
                ms.setProperty(k, 1, category="permissions")
        return ms

    def getEmails(self, role=None, perms=None, master_perm=None):
        emails = []
        members = self.getMembers(role=role, perms=perms, as_member=True)
        for m in members:
            if "invalid" in m.email:
                continue
            emails.append(m.email)
        if master_perm:
            emails = emails + Member.GetWithPermission(master_perm, email_list=True)
        return emails

    def sendEmail(self, role=None, perms=None, subject="Notification", template="email/base", body="", context={}, master_perm=None):
        c = {
            'settings': settings,
            'subject': subject,
            'from': settings.get("DEFAULT_FROM_EMAIL"),
            "body": body,
            'group': self,
            'sent_to': None,
        }
        sent_to = []
        c.update(context)
        members = self.getMembers(role=role, perms=perms, as_member=True)
        for m in members:
            if "invalid" in m.email:
                continue
            # print m.email
            c["to"] = m.email
            sent_to.append(m.email)
            c["user"] = m
            rest_mail.render_to_mail(template, c)

        if master_perm:
            c["to"] = Member.GetWithPermission(master_perm, email_list=True)
            c["sent_to"] = c["to"]
            if c["to"]:
                rest_mail.render_to_mail(template, c)

    def sendEvent(self, name, message, custom=None):
        if not custom:
            custom = {}
        custom["group_id"] = self.id
        RemoteEvents.sendToGroup(
            self,
            name,
            message=message,
            custom=custom)

    def sendChangeEvent(self, model, model_pk, name="group.change", custom=None):
        if not custom:
            custom = {}
        custom["group_id"] = self.id
        RemoteEvents.sendToGroup(
            self,
            name,
            model=model,
            model_id=model_pk,
            custom=custom)

    def getStats(self):
        return {
            "members": self.memberships.count(),
            "active": self.memberships.filter(state__gte=0).count(),
            "pending_invites": self.memberships.filter(state__in=[-10,-9]).count()
        }

    def logEvent(self, kind, component=None, component_id=None, note=None):
        GroupFeed = RestModel.getModel("account", "GroupFeed")
        member = None
        request = self.getActiveRequest()
        if request:
            member = request.member
        return GroupFeed.log(member, self, kind, component, component_id, note)

    def __str__(self):
        return F"{self.name}:{self.id}"

    @classmethod
    def canSubscribeTo(cls, credentials, msg):
        # this is called by the ws4redis framework when a websocket is requesting a subscription
        # credentials are the authentication details for the websocket
        # the response is expected to be a list of primary keys these credentials have access to
        # or the single pk if msg.pk is not none if the creds have access to that key
        if credentials.kind == "member":
            if msg.pk is not None:
                if credentials.instance.hasPermission("view_all_groups") or credentials.instance.isMemberOf(msg.pk):
                    return [msg.pk]
                return None
            return credentials.instance.getGroupIDs()
        return None

    @classmethod
    def canPublishTo(cls, credentials, msg):
        # this is called by the ws4redis framework when a websocket wants to publish to a channel
        # credentials are the authentication details for the websocket
        # this should return true or false
        if credentials.kind == "member":
            if msg.pk is not None:
                return credentials.instance.isMemberOf(msg.pk)
        return False         


class GroupMetaData(MetaDataBase):
    parent = models.ForeignKey(Group, related_name="properties", on_delete=models.CASCADE)







import PIL
from PIL import Image, ImageChops
from PIL.GifImagePlugin import getheader, getdata
try:
    from PIL.GifImagePlugin import _write_multiple_frames
except:
    _write_multiple_frames = None

def intToBin(i):
    """ Integer to two bytes """
    i1 = i % 256
    i2 = int( i/256)
    return chr(i1) + chr(i2)


def getheaderAnim(im):
    """ Animation header. To replace the getheader()[0] """
    bb = "GIF89a"
    bb += intToBin(im.size[0])
    bb += intToBin(im.size[1])
    bb += "\x87\x00\x00"
    return bb


def getAppExt(loops=0):
    """ Application extention. Part that secifies amount of loops. 
    if loops is 0, if goes on infinitely.
    """
    bb = "\x21\xFF\x0B"  # application extension
    bb += "NETSCAPE2.0"
    bb += "\x03\x01"
    if loops == 0:
        loops = 2**16-1
    bb += intToBin(loops)
    bb += '\x00'  # end
    return bb


def getGraphicsControlExt(duration=0.1):
    """ Graphics Control Extension. A sort of header at the start of
    each image. Specifies transparancy and duration. """
    bb = '\x21\xF9\x04'
    bb += '\x08'  # no transparancy
    bb += intToBin( int(duration*100) ) # in 100th of seconds
    bb += '\x00'  # no transparant color
    bb += '\x00'  # end
    return bb

# def run(fp, images, durations, loops):
# 	pass

def getPalette(images):
    # check which colors are used
    used_palette_colors = []
    for im in images:
        for i, count in enumerate(im.histogram()):
            if count and i not in used_palette_colors:
                used_palette_colors.append(i)
    used_palette_colors.sort()
    return used_palette_colors

def run(fp, images, durations, loops):
    """
    Given a set of images writes the bytes to the specified stream.
    """

    if hasattr(durations, '__len__'):
        if len(durations) != len(images):
            raise ValueError("number of durations mismatch with infiles")
    else:
        durations = [durations] * len(images)

    img = images[0]
    img.save(fp, save_all=True, append_images=images[1:], duration=durations, loop=0)
    return len(images)

def writeGif(fp, images, durations, loops):
    _write_multiple_frames(im, fp, images)

def writeLegacyGif(fp, images, durations, loops):
    frames = 0
    previous = None
    cimages = []
    for img in images:
        cimg = img.convert('P', palette=Image.ADAPTIVE, dither=1)
        cimages.append(cimg)

    palette = getPalette(cimages)
    print(("pallette size is: {}".format(len(palette))))
    for img in cimages:
        if not previous:
            # first image
            # gather data
            data = getdata(img)
            imdes, data = data[0], data[1:]
            header = getheaderAnim(img)
            appext = getAppExt(loops)
            graphext = getGraphicsControlExt(durations[0])

            # write global header
            fp.write(header)
            fp.write(palette)
            fp.write(appext)

            # write image
            fp.write(graphext)
            fp.write(imdes)
            for d in data:
                fp.write(d)
        else:
            # gather info (compress difference)
            data = getdata(img) 
            imdes, data = data[0], data[1:]
            graphext = getGraphicsControlExt(durations[frames])

            # write image
            fp.write(graphext)
            fp.write(imdes)
            for d in data:
                fp.write(d)
        # previous = img.copy()
        previous = True
        frames = frames + 1
    fp.write(";")  # end gif
    return frames




from rest import decorators as rd
from rest import views as rv
from rest import settings
from rest import helpers as rh
from taskqueue.models import Task
from pushit import utils
import time


@rd.urlPOST('hooks/git_update')
def rest_on_git_hook(request):
    sec_key = request.DATA.get("token")
    git_key = settings.get("GIT_KEY", "hookswhat")
    hook_request = None

    req_key = request.DATA.getHeader("HTTP_X_GITLAB_TOKEN")
    if req_key is not None:
        hook_request = utils.parseGitLab(request)
    else:
        req_key = request.DATA.getHeader("HTTP_X_HUB_SIGNATURE_256")
        if req_key is not None:
            hook_request = utils.parseGithub(request)    

    if sec_key != git_key and req_key is None:
        rh.error("GIT HOOK NO TOKEN", request.META)
        return rv.restPermissionDenied(request)

    proj_info = utils.getProjectInfo(hook_request.name, hook_request.branch)
    if proj_info is None:
        return rv.restStatus(request, False, error=f"no local project found {hook_request.name}:{hook_request.branch}")

    if hook_request.kind == "push":
        on_git_push_request(proj_info, hook_request)

    return rv.restStatus(request, True)


def on_git_push_request(info, hook_request):
    branch = info.get("branch")
    use_tq = info.get("use_tq", False)

    if use_tq:
        data = dict(branch=branch, project=hook_request.name, kind=hook_request.kind)
        Task.Publish("pushit", "run_update", data, channel="tq_updater")
    else:
        cmd = info.get("updater")
        asyncUpdater(cmd, branch)


@rd.rest_async
def asyncUpdater(cmd, branch):
    import random
    # randomize the hit so we avoid collisions (gitlabs issues)
    time.sleep(random.randint(2, 15))
    utils.updateCode(cmd, branch)




# from .models import *
from rest import crypto
# from rest.middleware import get_request
from rest import helpers
from rest.crypto import util
from rest.uberdict import UberDict
from rest import settings
from . import csv as rcsv
from . import excel as rxsl
from version import VERSION

from django.db import models
from django.http import HttpResponse, StreamingHttpResponse, Http404
from django.shortcuts import render
from django.db import connection
from django.db.models.query import QuerySet
from django.utils.datastructures import MultiValueDict
# from account.models import User
from django.core.exceptions import FieldDoesNotExist
from django.db.models import Q
from django.core.files.uploadedfile import UploadedFile
from io import StringIO

from auditlog.models import PersistentLog

from decimal import Decimal
import math
import datetime
import os
import sys
import types
import time
import pprint
import re
import inspect
import base64
import copy
import json
from . import profiler

CHUNKDIR = os.path.join(os.environ.get('TMPDIR', '/tmp'), 'chunked_uploads')


def toJSON(data, **kwargs):
    return json.dumps(data, cls=JSONEncoderExt)
    # return json.dumps(data, **kwargs)

def prettyJSON(data):
    return json.dumps(data, cls=JSONEncoderExt, sort_keys=True, indent=4, separators=(',', ': '))

try:
    import ujson
    def toJSON(data, **kwargs):
        # helpers.log_print(data)
        return ujson.dumps(data)

    # def prettyJSON(data):
    #   return json.dumps(data, sort_keys=True, indent=4, separators=(',', ': '))
except Exception:
    helpers.log_print("recommend installing ujson!")

# is this useless??? we have rest_serialize
class JSONEncoderExt(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime.datetime):
            return time.mktime(obj.timetuple())
        elif isinstance(obj, datetime.date):
            return obj.strftime("%Y/%m/%d")
        elif isinstance(obj, Decimal):
            if obj.is_nan():
                return 0.0
            return float(obj)
        elif isinstance(obj, float):
            if math.isnan(obj):
                return 0.0
        elif isinstance(obj, set):
            # helpers.log_error(obj)
            return str(obj)
        elif isinstance(obj, bytes):
            return util.toString(obj)
        try:
            return super().default(obj)
        except Exception:
            pass
        return "not parsable"


class RawQuery(object):
    def __init__(self, model, distinct=False, returns=None, joins=None, where=None, groupby=None, having=None, sort=None, limit=None, offset=None, count_by=None, last_sort=None, uniques=[]):
        self.model = model
        self.distinct = distinct
        self.returns = returns or []
        self.joins = joins or []
        self.where = where or []
        self.groupby = groupby or []
        self.having = having or []
        self.sort = sort
        self.limit = limit
        self.offset = offset
        self.count_by = count_by
        self.last_sort = last_sort
        self.uniques = uniques

    def addWhere(self, where):
        self.where.append(where)

    def addHaving(self, where):
        self.having.append(where)

    def addJoin(self, model, joinon, jointype=None, joinas=None):
        j = {'model': model, 'joinon': joinon}
        if jointype:
            j['jointype'] = jointype
        if joinas:
            j['joinas'] = joinas
        self.joins.append(j)

    def joined(self, model):
        for j in self.joins:
            if j['model'] == model:
                return True
        return False

    def count(self):
        count_by = self.count_by or '*'
        q = RawQuery(model=self.model, distinct=self.distinct, returns=["COUNT(%s)" % count_by], joins=self.joins, where=self.where, groupby=self.groupby, having=self.having)
        return q.run_raw()[0]

    def sortlist(self):
        if not type(self.sort) in (list,tuple):
            osort = [self.sort, self.last_sort]
        else:
            osort = list(self.sort) + [self.last_sort]
        try:
            osort.remove(None)
        except ValueError:
            pass
        return osort

    def reverse(self):
        osort = self.sortlist()
        self.sort = []
        for s in osort:
            if s[0] == '-':
                s = '+' + s[1:]
            elif s[0] == '+':
                s = '-' + s[1:]
            else:
                s = '-' + s
            self.sort.append(s)
        return self

    def serialize_datetime(self, when):
        if connection.vendor == "mysql":
            return 'FROM_UNIXTIME(%f)' % time.mktime(when.timetuple())
        elif connection.vendor == "postgresql":
            return "(TIMESTAMP WITH TIME ZONE 'epoch' + %f * INTERVAL '1 second')" % time.mktime(when.timetuple())
        elif connection.vendor == "sqlite":
            return "DATETIME(%i.%i, 'unixepoch', 'localtime')" % (time.mktime(when.timetuple()), getattr(when, 'microsecond', 0))
        else:
            return '%s' % time.strftime("%Y-%m-%d %H:%M:%S", when.timetuple())

    def quote_string(self, data):
        data = str(data)
        data = data.replace('\\', '\\\\')
        data = data.replace('"', '\\"')
        data = data.replace("'", "\\'")
        data = data.replace('%', '%%')
        return "'" + str(data) + "'"

    def serialize(self, data):
        if isinstance(data, Decimal):
            if data.is_nan():
                return "0"
            return str(data)
        elif isinstance(data, float):
            if math.isnan(data):
                return "0"
            return str(data)
        elif isinstance(data, int):
            return str(data)
        elif type(data) in (datetime.datetime, datetime.date):
            return self.serialize_datetime(data)
        else:
            return self.quote_string(data)

    def sql(self):
        sql = "SELECT"
        if self.distinct:
            sql += " DISTINCT"
        if len(self.returns) > 0:
            sql += " " + ",".join(self.returns)
        else:
            sql += " *"
        sql += " FROM %s" % self.model._meta.db_table
        for j in self.joins:
            sql += " %s %s AS %s ON (%s)" % (
                j.get('jointype', 'LEFT OUTER JOIN'),
                j['model']._meta.db_table,
                j.get('joinas', j['model']._meta.db_table),
                " AND ".join(j['joinon']),
            )
        if len(self.where) > 0:
            sql += " WHERE " + " AND ".join(self.where)
        if len(self.groupby) > 0:
            sql += " GROUP BY " + ",".join(self.groupby)
        if len(self.having) > 0:
            sql += " HAVING " + " AND ".join(self.having)
        if self.sort:
            sort = []
            for s in self.sortlist():
                add = ''
                if s[0] == '-':
                    s = s[1:]
                    add =' DESC'
                elif s[0] == '+':
                    s = s[1:]
                    add = ' ASC'
                sort.append(s + add)
                if s in self.uniques:
                    break
            sql += " ORDER BY " + ",".join(sort)
        if self.limit:
            sql += " LIMIT %d" % self.limit
        if self.offset:
            sql += " OFFSET %d" % self.offset

        return sql + ";"
    
    def run(self):
        return self.model.objects.raw(self.sql())

    def run_raw(self):
        cursor = connection.cursor()
        cursor.execute(self.sql())
        return cursor.fetchone()


def _getFields(qset=None, model=None):
    """
    returns a list of fields for specified QuerySet or Model
    """
    if hasattr(qset, '_meta') and getattr(qset, '_meta') != None:
        fields = qset._meta.fields
    elif hasattr(qset, 'model') and getattr(qset, 'model') != None:
        fields = qset.model._meta.fields
    elif hasattr(qset, 'keys'):
        return list(qset.keys())
    elif hasattr(model, '_meta') and getattr(model, '_meta') != None:
        fields = model._meta.fields
    else:
        return None
    if fields:
        return list(f.name for f in fields)
    return None


def requestInPaths(request, paths):
    for ep in paths:
        if request.path.startswith(ep):
            return True
    return False


def _returnResults(request, data, accept_list=None, status=200):
    return restResult(request, data, accept_list, status)


def restResult(request, ret, accept_list=None, status=200):
    """
    returns a response data object in json, text or html
    """
    # figure return type
    if not accept_list:
        if request and request.DATA.get('_type', None):
            accept_list = [request.DATA.get('_type')]
        elif request and 'HTTP_ACCEPT' in request.META:
            accept_list = request.META["HTTP_ACCEPT"].split(',')
        elif request and 'HTTP_ACCEPT_ENCODING' in request.META:
            accept_list = request.META["HTTP_ACCEPT_ENCODING"].split(',')
        else:
            accept_list = []

    for i in range(len(accept_list)):
        a2 = accept_list[i].split(";", 1)
        if len(a2) == 2:
            accept_list[i] = a2[0]

    if "data" not in accept_list:
        ret["datetime"] = time.time()
        ret["elapsed"] = int((time.perf_counter() - request._started)*1000)
        if "status" not in ret:
            ret["status"] = True

    # render response
    if 'data' not in accept_list and hasattr(request, "rest_debug") and request.rest_debug:
        if settings.DEBUG_REST_NO_LISTS and "count" in ret:
            if "count" in ret:
                PersistentLog.log(dict(count=ret.get("count"), size=ret.get("size")), 0, request, "rest", action="response")
            else:
                PersistentLog.log(ret, 0, request, "rest", action="response")
        else:
            PersistentLog.log(ret, 0, request, "rest", action="response")

    if 'application/json' in accept_list:
        callback = request.DATA.get('callback', None)
        if callback:
            return HttpResponse("%s(%s)" % (callback, toJSON(ret)), content_type="application/json", status=status)
        try:
            return HttpResponse(toJSON(ret), content_type="application/json", status=status)
        except Exception as err:
            helpers.log_print(ret)
            raise err
    elif 'text/html' in accept_list:
        output = prettyJSON(ret)
        req_dict = request.DATA.asDict()
        req_out = prettyJSON(req_dict)
        # output = json.dumps(ret, sort_keys=True, indent=4, separators=(',', ': '), cls=JSONEncoderExt)
        # req_out = json.dumps(request.DATA.asDict(), sort_keys=True, indent=4, separators=(',', ': '), cls=JSONEncoderExt)

        # req_out = pprint.pformat(request.DATA.asDict(), indent=4).strip()
        path = request.path
        parts = path.split('/')
        last_pk = parts.pop()
        if last_pk and last_pk.isdigit():
            path = "/".join(parts)
        else:
            last_pk = ""

        is_debug = getattr(settings, "DEBUG", False) or getattr(settings, "REST_DEBUGGER", False)
        rest_help = {}
        if is_debug and hasattr(request, "rest_class"):
            rest_help = request.rest_class.get_rest_help()
            rest_help = prettyJSON(rest_help)
        context = {
            "help": rest_help,
            "input": req_dict,
            "pk": last_pk,
            "output_dict": ret,
            "output": output,
            "req_out":req_out,
            "path":path,
            "method": request.method,
            "debug":is_debug,
            "request": request,
            "version": VERSION
        }
        return render(request, "rest_html.html", context, status=status)
    elif 'text/plain' in accept_list:
        return HttpResponse(pprint.pformat(ret), content_type="text/plain", status=status)
    elif 'data' in accept_list:
        return ret
    else:
        return HttpResponse(toJSON(ret), content_type="text/plain", status=status)

def _filter_recurse(remove, lst):
    if isinstance(lst, dict):
        ret = {}
    else:
        ret = []
    for item in lst:
        if isinstance(lst, dict):
            if item[:len(remove)+1] == remove + ".":
                ret[item[len(remove)+1:]] = lst[item]
        elif type(item) in (list,tuple) and type(item[0]) in (str, str):
            p = item[0]
            if p[:len(remove)+1] == remove + ".":
                ret.append((p[len(remove)+1:], item[1]))
        elif isinstance(item, str):
            if item[:len(remove)+1] == remove + ".":
                ret.append(item[len(remove)+1:])

    return ret

def restReturn(request, data, accept_list=None):
    return restResult(request, data, accept_list)

def __call_func(func, *args, **kwargs):
    if not kwargs:
        return func(*args)
    if hasattr(inspect, "getargspec"):
        # removed in python3.11 (but much faster)
        take = inspect.getargspec(func)[0]
    else:
        take = list(inspect.signature(func).parameters.keys())
    give = {}
    for arg in kwargs:
        if arg in take:
            give[arg] = kwargs[arg]
    return func(*args, **give)

def rest_serialize(data):
    if isinstance(data, datetime.datetime):
        try:
            return time.mktime(data.timetuple())
        except:
            print(("invalid date: {0}".format(data)))
        return 0.0
    elif isinstance(data, datetime.date):
        return data.strftime("%Y/%m/%d")
    elif isinstance(data, dict):
        data = data.copy()
        for k in list(data.keys()):
            data[k] = rest_serialize(data[k])
    elif isinstance(data, list) or isinstance(data, tuple):
        newdata = []
        for v in data:
            newdata.append(rest_serialize(v))
        data = newdata
    elif isinstance(data, float):
        if math.isnan(data):
            return 0.0
    elif isinstance(data, Decimal):
        if data.is_nan():
            return 0.0
        return float(data)
    return data

def updateFeature(func, request, qset, featuresets={}, features=[], **kwargs):
    features = list(features)
    kwargs = kwargs.copy()
    if not featuresets:
        return None

    got_features = []
    for f in getattr(request, 'REQUEST', {}).get('features', '').split(','):
        fs = f.strip()
        if fs != '':
            features.append(fs)

    agent = getattr(request, 'META', {}).get('HTTP_USER_AGENT', '')
    for a in getattr(settings, 'REST_USERAGENT_FEATURES', []):
        if re.match(a[0], agent):
            for f in a[1]:
                features.append(f)

    for f in features:
        if f in got_features:
            continue
        if f in featuresets and not f in got_features:
            got_features.append(f)
            for p,v in list(featuresets[f].items()):
                p_type = 'A'
                if p[:4] == 'DEL_':
                    p_type = 'D'
                    p = p[4:]
                elif p[:4] == 'ADD_':
                    p_type = 'A'
                    p = p[4:]
                elif p[:8] == 'REPLACE_':
                    p_type = 'R'
                    p = p[8:]

                if p_type == 'A' and p == 'features':
                    for ff in v:
                        features.append(ff)
                    continue

                if not p in kwargs:
                    raise ValueError('Invalid feature key: %s' % p)

                if p_type == 'R':
                    kwargs[p] = copy.deepcopy(v)
                    continue

                if type(kwargs[p]) == tuple:
                    kwargs[p] = list(kwargs[p])
                else:
                    kwargs[p] = copy.deepcopy(kwargs[p])

                if p_type == 'D' and hasattr(kwargs[p], 'remove'):
                    for i in v:
                        try:
                            kwargs[p].remove(i)
                        except ValueError:
                            pass
                elif p_type == 'D' and hasattr(kwargs[p], '__delitem__'):
                    for i in v:
                        try:
                            del(kwargs[p][i])
                        except KeyError:
                            pass
                elif p_type == 'A' and hasattr(kwargs[p], 'append'):
                    for i in v:
                        if not i in kwargs[p]:
                            kwargs[p].append(i)
                elif p_type == 'A' and hasattr(kwargs[p], '__setitem__'):
                    for i in v:
                        kwargs[p][i] = v[i]
                else:
                    raise ValueError('Invalid feature action: %s' % p)

    if got_features:
        return func(request, qset, features=features, **kwargs)
    else:
        return None


@profiler.timeit
def restGet(request, qset, model=None, fields=None, extra=[], exclude=[], fkey_depth=0, recurse_into=[], filter={}, orig_request=None, orig_objs={}, ignore_noattr=False, require_perms=[], accept_list=None, featuresets={}, features=[], return_httpresponse=True):
    """
    request: HttpRequest object
    qset: QuerySet or data object to return
    model: (optional) model of the queryset
    fields: (optional) fields to return, default to all fields.  string, or tuple(db_field, output_field)
    extra: (optional) extra fields to return
    exclude: (optional) do not display these fields
    fkey_depth: (optional) levels to recurse into foreign keys, default 0
    recurse_into: (optional) list of foreign keys in data that should be recursed into, regardless of fkey_depth
    filter: (optional) filters for recurse fetch (default all)
    ignore_noattr: (optional) ignore unknown attributes (default false)
    accept_list: (optional) requested return format
    RETURN: HttpResponse

    process a RESTful GET request to get an object.  User is responsible for checking permissions, etc before calling.
    """
    # if not request:
    #   request = get_request()

    # ret = updateFeature(restGet, request, qset, model=model, fields=fields, extra=extra, exclude=exclude, fkey_depth=fkey_depth, recurse_into=recurse_into, filter=filter, orig_request=orig_request, orig_objs=orig_objs, ignore_noattr=ignore_noattr, require_perms=require_perms, accept_list=accept_list, featuresets=featuresets, features=features)
    # if ret != None:
    #     return ret

    # for p in require_perms:
    #     if not (request.user and request.user.has_perm(p)):
    #         return restResult(request, {'status': False, 'error': 'Permission denied'}, accept_list=accept_list)

    if not orig_request:
        orig_request=request
    ret = {}

    # if no qset then 404
    if not qset:
        raise Http404
    if type(qset) == QuerySet:
        try:
            qset = qset[0]
        except IndexError:
            raise Http404

    if not fields or list(fields) == ["*"]:
        fields = _getFields(qset, model)
        #print fields
    if not model:
        model = getattr(qset, "__class__", None)

    if request and model:
        if hasattr(model, "on_rest_request") and not hasattr(request, "rest_class"):
            request.rest_class = model

    objs = orig_objs.copy()
    objs['obj__parent'] = qset
    if model and hasattr(model, '__name__'):
        objs['obj_' + model.__name__] = qset

    if "*" in exclude:
        fields = []

    for f in extra:
        if f[:2] == "*.":
            f = f[2:]
        if "." in f:
            continue
        if not f in fields:
            fields.append(f)
    # helpers.log_print("fields: {}".format(fields))
    recursed = []
    # get fields into data object, recurse if necessary
    for f in fields:
        if type(f) in (list, tuple):
            fout = f[1]
            f = f[0]
        else:
            fout = f

        fargs = []
        if type(f) is str and "|" in f:
            fargs = f.split("|")
            f = fargs.pop(0)
        if type(f) in (types.FunctionType, types.LambdaType):
            data = f(qset)
        elif f in exclude or "*."+f in exclude:
            continue
        elif f[0] == "!":
            f = f[1:]
            if fout[0] == "!":
                fout = fout[1:]
            data = qset
            for s in f.split("."):
                data = getattr(data, s)
                if data == None:
                    break
                if hasattr(data, '__func__'):
                    data = __call_func(data, request = orig_request, obj__self=qset, **orig_objs)
        else:
            if "." in f:
                continue
            try:
                # exc_info = sys.exc_info()
                # traceback.print_stack()
                data = getattr(qset, f)
            except AttributeError as err:
                try:
                    data = qset[f]
                except (TypeError, KeyError, IndexError):
                    if ignore_noattr:
                        continue
                    else:
                        # traceback.print_stack()
                        # traceback.print_exception(*exc_info)
                        # print str(traceback.format_exc())
                        helpers.log_print("{} has no attribute: {}".format(type(qset), f), fields)
                        # helpers.log_exception(err)
                        continue
        if isinstance(data, dict) and hasattr(data, "asDict"):
            data = data.asDict()
        elif hasattr(data, '__func__'):
            # helpers.log_error("__func__", repr(f), repr(type(data)), repr(data))
            data = __call_func(data, *fargs, request=orig_request, obj__self=qset, **orig_objs)
        elif f.startswith("get_") and f.endswith("_display") and callable(data):
            data = data()

        # helpers.log_print("{} ({}) {}".format(f, type(data), data))
        data = rest_serialize(data)
        # print("{} ({}) {}".format(f, type(data), data))
        if isinstance(data, models.Model):
            if f in recurse_into:
                pass
            elif fkey_depth > 0:
                recurse_into.append((f, fout))
            elif hasattr(qset, f + "_id"):
                ret[fout] = getattr(qset, f + "_id")
            else:
                ret[fout] = data.pk
        elif hasattr(data, 'all'):
            if f in recurse_into:
                pass
            elif fkey_depth > 0:
                recurse_into.append((f, fout))
        else:
            ret[fout] = data


    # helpers.log_print("fields", fields)
    # helpers.log_print("recurse_into", recurse_into)

    for f in recurse_into:
        # print "-"*30
        # print recurse_into
        if '.' in f: continue
        if type(f) in (list, tuple):
            fout = f[1]
            f = f[0]
        else:
            fout = f
        fname = fout
        if fname == "":
            fname = f
        if f[:2] == "*.":
            f = f[2:]
        if fname in recursed:
            continue
        recursed.append(fname)
        rset = None
        odata = {}
        has_field = hasattr(qset, f)

        if f.startswith("generic__") and hasattr(qset, "restGetGenericRelation"):
            # print "handling generic: " + f
            # our generic relationship parser
            generic_graph = "generic"
            if fout != f:
                generic_graph = fout
            fout = f.split("__")[1]
            try:
                obj = qset.restGetGenericRelation(fout)
            except Exception as err:
                helpers.log_print("generic relationship error")
                helpers.log_exception(err)
                obj = None
            if not obj:
                continue
            # print ret
            odata = restGet(request, obj, accept_list=['data'], orig_request=orig_request, **obj.getGraph(generic_graph))
            odata["model"] = getattr(qset, fout)
            gkey = "{0}_id".format(fout)
            if gkey in ret:
                del ret[gkey]
        elif has_field and (hasattr(getattr(qset, f), 'all') or isinstance(getattr(qset, f), list)):
            rset = getattr(qset, f)
            odata = []
        elif isinstance(qset, dict) and f in qset and (hasattr(qset.get(f), 'all') or isinstance(qset.get(f), list)):
            rset = qset[f]
            odata = []
        elif hasattr(qset, f + '_set'):
            rset = getattr(qset, f + '_set')
            odata = []
        elif has_field and isinstance(getattr(qset, f), models.Model):
            odata = restGet(request, getattr(qset, f), fields=_filter_recurse(fname, fields), extra=_filter_recurse(fname, extra), exclude=_filter_recurse(fname, exclude), fkey_depth=fkey_depth-1, recurse_into=_filter_recurse(fname, recurse_into), filter=_filter_recurse(fname, filter), accept_list=['data'], orig_request=orig_request, orig_objs=objs)
        elif isinstance(qset, dict) and f in qset and isinstance(qset.get(f), models.Model):
            odata = restGet(request, qset.get(f), fields=_filter_recurse(fname, fields), extra=_filter_recurse(fname, extra), exclude=_filter_recurse(fname, exclude), fkey_depth=fkey_depth-1, recurse_into=_filter_recurse(fname, recurse_into), filter=_filter_recurse(fname, filter), accept_list=['data'], orig_request=orig_request, orig_objs=objs)
        elif isinstance(qset, dict) and f in qset and isinstance(qset.get(f), dict):
            odata = qset.get(f)
        elif has_field and callable(getattr(qset, f)):
            rset = getattr(qset, f)()
            if rset is not None:
                odata = []
        # elif has_field and hasattr(getattr(qset, f), 'im_func'):
        #     rset = getattr(qset, f)()
        #     if rset is not None:
        #         odata = []
        elif isinstance(qset, dict) and f in qset and callable((qset.get(f))):
            rset = qset.get(f)()
            odata = []
        else:
            odata = {}
        if rset and isinstance(rset, models.Model):
            odata = restGet(request, rset, fields=_filter_recurse(fname, fields), extra=_filter_recurse(fname, extra), exclude=_filter_recurse(fname, exclude), fkey_depth=fkey_depth-1, recurse_into=_filter_recurse(fname, recurse_into), filter=_filter_recurse(fname, filter), accept_list=['data'], orig_request=orig_request, orig_objs=objs)
            rset = None
        if rset:
            filt = filter.get(fname, None)
            if isinstance(rset, list):
                pass
            elif filt:
                rset = rset.filter(**filt)
            elif hasattr(rset, "all"):
                rset = rset.all()

            if type(rset) is list and len(rset) and not hasattr(rset[0], "id"):
                odata = rset
            elif isinstance(rset, dict) or type(rset) in [float, int, str]:
                odata = rset
            else:
                # print "-" * 80
                # print fname
                # print fields
                odata = restListEx(None, rset, size=0, fields=_filter_recurse(fname, fields), extra=_filter_recurse(fname, extra), exclude=_filter_recurse(fname, exclude), fkey_depth=fkey_depth-1, recurse_into=_filter_recurse(fname, recurse_into), filter=_filter_recurse(fname, filter), accept_list=['data'], orig_request=orig_request, orig_objs=objs)

        if fout:
            #print "1: {0} = {1}".format(fout, odata);
            if isinstance(odata, dict) and len(odata) == 0:
                odata = None
            ret[fout] = odata
        else:
            # print "2: {0} = {1}".format(fout, odata);
            if type(odata) == list:
                if len(odata[:1]) == 0:
                    continue
                odata = odata[0]
            for x in odata:
                # never allow overriding id for emptry recurse
                if x == "id":
                    continue
                ret[x] = odata[x]

    if return_httpresponse:
        if not (accept_list != None and 'data' in accept_list):
            ret = { 'status': True, 'data': ret }
        return restResult(request, ret, accept_list)

    return ret

def _order_list(ordering, data, dir, clear=False):
    ret = {"_dir": dir}
    if dir == '=':
        for r in data:
            ret[r] = data[r]
    else:
        for (o, rev) in ordering:
            try:
                name = o.attname
            except AttributeError:
                name = o.split(".")[-1]
            r = getattr(data, name)
            if type(r) in (datetime.datetime, datetime.date):
                r = {"datetime": [int(time.mktime(r.timetuple())), getattr(r, 'microsecond', 0)]}
            ret[getattr(o, 'name', o)] = r
    ret = toJSON(ret);
    if not clear:
        ret = base64.b64encode(ret.encode("utf-8")).decode("utf-8")
    return ret


def restListOther(request, qset, size=25, start=0, sort=None, accept_list=None, return_httpresponse=True):

    if request:
        if sort is None:
            sort = request.DATA.get("sort", sort)
        size = request.DATA.get(['size', '_size'], size, field_type=int)
        start = request.DATA.get(['start', '_start'], start, field_type=int)

    # if sort and sort.lstrip('-') in fields:
    #   sort = rsort

    if sort:
        reversed = sort[0] == '-'
        prop = sort
        if reversed:
            prop = sort.lstrip('-')

        qset.sort(key=lambda props: getattr(props, prop), reverse=reversed)

    data_list = []
    for obj in qset:
        if type(obj) in [float, int, str, Decimal]:
            data_list.append(obj)
        elif isinstance(obj, dict):
            data_list.append(rest_serialize(obj))
        elif hasattr(obj, "restGet"):
            data_list.append(obj.restGet(request, as_dict=True))
        else:
            helpers.log_print("unhandled type: {}: '{}'".format(type(obj), obj))

    if not return_httpresponse:
        return data_list
    count = len(data_list)
    data_list = data_list[:size]
    # build response
    ret = {}
    ret["size"] = size
    ret["count"] = count
    ret["start"] = start
    ret["data"] = data_list
    return restResult(request, ret, accept_list)


@profiler.timeit
def restList(request, qset, model=None, size=25, start=0, sort=None, fields=None, extra=[], exclude=[], fkey_depth=0, recurse_into=[], filter={}, get_count=True, orig_request=None, orig_objs={}, page=None, ignore_noattr=False, todata=lambda x: x, require_perms=[], accept_list=None, featuresets={}, features=[], return_httpresponse=True, totals=None):
    """
    request: HttpRequest object
    qset: QuerySet or data object to return
    model: (optional) model of the queryset
    size: (optional) return set size, default 25, can override via GET[size].  0=return all.
    start: (optional) return set offset, default 0, can override via GET[start]
    sort: (optional) return set sorting, default None, can override via GET[sort]
    fields: (optional) fields to return, default to all fields
    extra: (optional) extra fields to return
    exclude: (optional) do not display these fields
    fkey_depth: (optional) levels to recurse into foreign keys, default 0
    recurse_into: (optional) list of foreign keys in data that should be recursed into, regardless of fkey_depth
    filter: (optional) filters for recurse fetch (default all)
    get_count: (optional) boolean of whether to get a count of total records available
    page: (optional) pagination token, can override via GET[page]
    ignore_noattr: (optional) ignore unknown attributes (default false)
    todata: (optional) function to convert per-element data into data for restGet
    accept_list: (optional) requested return format
    RETURN: HttpResponse

    process a RESTful GET request to list a set of objects.  User is responsible for checking permissions, etc before calling.
    """
    ffunc = updateFeature(restList, request, qset, model=model, size=size, start=start, sort=sort, fields=fields, extra=extra, exclude=exclude, fkey_depth=fkey_depth, recurse_into=recurse_into, filter=filter, get_count=get_count, orig_request=orig_request, orig_objs=orig_objs, page=page, ignore_noattr=ignore_noattr, todata=todata, require_perms=require_perms, accept_list=accept_list, featuresets=featuresets, features=features)
    if ffunc != None:
        return ffunc

    if type(qset) in [list, tuple]:
        return restListOther(request, qset, size, start, sort, accept_list, return_httpresponse)
    elif type(qset) != QuerySet:
        raise Exception("qset of type '{}' not supported for restList".format(type(qset)))

    if not model:
        model = getattr(qset, "model", None)

    if not fields:
        fields = _getFields(qset, model)

    if request:
        # values from request
        size = request.DATA.get(['size', '_size'], size, field_type=int)
        start = request.DATA.get(['start', '_start'], start, field_type=int)
        if sort is None:
            sort = request.DATA.get('sort', sort)

    count = 0
    if get_count:
        count = qset.count()

    if sort and isinstance(sort, str):
        if "metadata" in sort:
            sort = ""
        elif sort.endswith("_display"):
            # fix for django _display kinds being sorted
            sort = sort[:sort.find("_display")]
    # handle sorting
    sort_args = None
    if sort and type(qset) == QuerySet:
        sort_args = []
        for s in sort.split(","):
            s = s.replace('.', '__')
            sort_args.append(s)

        if sort_args:
            try:
                qset = qset.order_by(*sort_args)
            except Exception:
                helpers.log_exception("sorting error", sort_args)

    qset, ret = updatePagination(request, qset, size, count, start, page)

    qset = qset[:size]
    # print fields
    data_list = []
    for obj in qset:
        if type(obj) in [list, tuple, QuerySet]:
            data_list.append(restList(request, todata(obj), model=model, fields=fields, extra=extra, exclude=exclude, fkey_depth=fkey_depth, recurse_into=recurse_into, filter=filter, ignore_noattr=ignore_noattr, accept_list=['data'], orig_request=orig_request, orig_objs=orig_objs, return_httpresponse=False))
        else:
            data_list.append(restGet(request, todata(obj), model=model, fields=fields, extra=extra, exclude=exclude, fkey_depth=fkey_depth, recurse_into=recurse_into, filter=filter, ignore_noattr=ignore_noattr, accept_list=['data'], orig_request=orig_request, orig_objs=orig_objs))

    if not return_httpresponse:
        return data_list

    if totals:
        ret.update(totals)

    # build response
    ret["size"] = size
    ret["count"] = count
    ret["start"] = start
    ret["data"] = data_list
    if request is not None and hasattr(request, "response_extra"):
        ret.update(request.response_extra)
    if sort_args:
        ret["sort"] = sort_args
    return restResult(request, ret, accept_list)


def updatePagination(request, qset, size, count, start, page):
    # decode our page data object if it exists
    if not request:
        return qset, {}
    page = request.DATA.get('page', page)

    if page:
        page_data = json.loads(base64.b64decode(page))
    else:
        page_data = {'_dir': '!'}

    if page_data['_dir'] == '=':
        start = page_data.get('start', start)

    ordering = None
    if type(qset) == QuerySet and qset.query.can_filter():
        ordering = []
        if not qset.ordered:
            modelorder = []
        elif qset.query.extra_order_by:
            modelorder = qset.query.extra_order_by
        elif not qset.query.default_ordering:
            modelorder = qset.query.order_by
        else:
            modelorder = qset.query.order_by or qset.query.model._meta.ordering or []
        modelorder = list(modelorder)
        modelorder.append('pk')
        qset = qset.order_by(*modelorder)
        try:
            for o in qset.query.order_by:
                rev = False
                if o[0] == '-':
                    rev = True
                    o = o[1:]
                if o == 'pk':
                    o = qset.model._meta.pk
                else:
                    o = qset.model._meta.get_field(o) # raises FieldDoesNotExist
                ordering.append((o, rev))
                if o.unique:
                    break
        except FieldDoesNotExist:
            pass

    # parse supplied paging info
    do_reverse = False
    first_page = start == 0
    last_page = False
    if ordering:
        qfilter = None
        qfilter2 = {}
        rfilter = []
        rfilter2 = []
        for (o, rev) in ordering:
            name = getattr(o, 'name', o)
            if not name in page_data:
                continue
            # pull page data by name
            if isinstance(page_data[name], dict) and "datetime" in page_data[name]:
                data = datetime.datetime.fromtimestamp(page_data[name]["datetime"][0])
                data = data.replace(microsecond=page_data[name]["datetime"][1])
            elif isinstance(page_data[name], dict) and "date" in page_data[name]:
                data = datetime.date.fromtimestamp(page_data[name]["date"][0])
            else:
                data = page_data[name]

            if type(qset) == QuerySet:
                if rev == (page['_dir'] == '+'):
                    att = name + "__lt"
                else:
                    att = name + "__gt"

                qfilter3 = qfilter2.copy()
                qfilter3[att] = data
                if not qfilter:
                    qfilter = Q(**qfilter3)
                else:
                    qfilter = qfilter | Q(**qfilter3)

                if o.unique:
                    break

                qfilter2[o.name] = data

        if qfilter:
            qset = qset.filter(qfilter)
        if page_data['_dir'] == '-':
            do_reverse = True
        first_page = False

    # if do_reverse:
    #   if size:
    #       qset = qset.reverse()[:size+1]
    #   else:
    #       qset.reverse()

    if size:
        qset = qset[start:start+size+1]

    ret = {}
    ret["start"] = start

    if ordering:
        try:
            if not first_page:
                ret['page_prev'] = _order_list(ordering, qset[0], '-')
            else:
                ret['page_before'] = _order_list(ordering, qset[0], '-')
        except IndexError:
            pass
        try:
            if do_reverse:
                offset = -1
            else:
                offset = size-1
            if size > 0 and not last_page:
                ret['page_next'] = _order_list(ordering, qset[offset], '+')
            elif size > 0:
                ret['page_after'] = _order_list(ordering, qset[offset], '+')
        except IndexError:
            try:
                ret['page_after'] = _order_list(ordering, qset[-1], '+')
            except:
                pass

    return qset, ret

# def restListQuerySet(request, )

def restListEx(request, qset, model=None, size=25, start=0, sort=None, fields=None, extra=[], exclude=[], fkey_depth=0, recurse_into=[], filter={}, get_count=True, orig_request=None, orig_objs={}, page=None, ignore_noattr=False, todata=lambda x: x, require_perms=[], accept_list=None, featuresets={}, features=[], return_httpresponse=True):
    """
    request: HttpRequest object
    qset: QuerySet or data object to return
    model: (optional) model of the queryset
    size: (optional) return set size, default 25, can override via GET[size].  0=return all.
    start: (optional) return set offset, default 0, can override via GET[start]
    sort: (optional) return set sorting, default None, can override via GET[sort]
    fields: (optional) fields to return, default to all fields
    extra: (optional) extra fields to return
    exclude: (optional) do not display these fields
    fkey_depth: (optional) levels to recurse into foreign keys, default 0
    recurse_into: (optional) list of foreign keys in data that should be recursed into, regardless of fkey_depth
    filter: (optional) filters for recurse fetch (default all)
    get_count: (optional) boolean of whether to get a count of total records available
    page: (optional) pagination token, can override via GET[page]
    ignore_noattr: (optional) ignore unknown attributes (default false)
    todata: (optional) function to convert per-element data into data for restGet
    accept_list: (optional) requested return format
    RETURN: HttpResponse

    process a RESTful GET request to list a set of objects.  User is responsible for checking permissions, etc before calling.
    """
    start_time = time.time()
    ret = updateFeature(restList, request, qset, model=model, size=size, start=start, sort=sort, fields=fields, extra=extra, exclude=exclude, fkey_depth=fkey_depth, recurse_into=recurse_into, filter=filter, get_count=get_count, orig_request=orig_request, orig_objs=orig_objs, page=page, ignore_noattr=ignore_noattr, todata=todata, require_perms=require_perms, accept_list=accept_list, featuresets=featuresets, features=features)
    # print "s1 {}".format(time.time() - start_time)
    # print "!! warning calling depercated restListEx"
    if ret != None:
        return ret

    for p in require_perms:
        if not (request.user and request.user.has_perm(p)):
            return restResult(request, {'status': False, 'error': 'Permission denied'}, accept_list=accept_list)

    # print "s2 {}".format(time.time() - start_time)
    if not orig_request:
        orig_request=request
    ret = {}
    rsort = sort
    t3 = time.time()
    if not fields:
        fields = _getFields(qset, model)
    # print "\tt3 {}".format(time.time() - t3)
    if not model:
        model = getattr(qset, "model", None)
    # print "\tt3 {}".format(time.time() - t3)
    # get overrides via HTTP
    if request:
        try:
            if size == '*':
                size = 0
            else:
                size = int(request.DATA.get('size', size))
        except ValueError:
            pass
        try:
            if start == '*':
                start = 0
            else:
                start = int(request.DATA.get(['start', '_start'], start))
        except ValueError:
            pass
        if sort == '*':
            sort = None
            rsort = None
        else:
            rsort = request.DATA.get('sort', sort)

        if get_count:
            if type(get_count) == int:
                ret['count'] = get_count
            elif type(qset) in (list, tuple):
                ret['count'] = len(qset)
            elif qset == None:
                ret['count'] = 0
            elif hasattr(qset, "count"):
                t4 = time.time()
                # # print qset.query
                ret['count'] = qset.count()
                # print "\tt4 {}".format(time.time() - t4)
        page = request.DATA.get('page', page)
    # print "s3 {}".format(time.time() - start_time)
    try:
        page = json.loads(base64.b64decode(page))
        if not page.get('_dir') in ('+', '-', '='):
            raise ValueError
    except (ValueError, TypeError):
        page = {'_dir': '!'}

    if page['_dir'] == '=':
        start = page.get('start', start)

    ret['start'] = start
    ret['size'] = size

    if isinstance(rsort, str):
        if "metadata." in rsort:
            rsort = ""
    # do sorting
    if rsort:
        if type(qset) == RawQuery:
            if re.match("[+-]?[a-zA-Z0-9._,]+", rsort):
                qset.sort = rsort.split(",")
        elif not type(qset) in (list, tuple):
            sort = []
            # for s in str(rsort).split(","):
            #   if s.lstrip('-') in  (type(x) in (list,tuple) and x[0] or x for x in list(fields) + list(recurse_into)):
            #       sort.append(s.replace('.', '__'))
            for s in str(rsort).split(","):
                s = s.replace('.', '__')
                sort.append(s)

            if sort:
                qset = qset.order_by(*sort)
                ret['sort'] = sort
        elif type(qset) == list:
            if rsort and rsort.lstrip('-') in fields:
                sort = rsort

            if sort:
                reversed = rsort[0] == '-'
                prop = rsort
                if reversed:
                    prop = rsort.lstrip('-')

                qset.sort(key=lambda props: getattr(props, prop), reverse=reversed)
    # print "s4 {}".format(time.time() - start_time)
    # sane paging - get ordering list
    ordering = []
    try:
        if type(qset) == QuerySet and qset.query.can_filter():
            if not qset.ordered:
                modelorder = []
            elif qset.query.extra_order_by:
                modelorder = qset.query.extra_order_by
            elif not qset.query.default_ordering:
                modelorder = qset.query.order_by
            else:
                modelorder = qset.query.order_by or qset.query.model._meta.ordering or []
            modelorder = list(modelorder)
            modelorder.append('pk')
            qset = qset.order_by(*modelorder)
            for o in qset.query.order_by:
                rev = False
                if o[0] == '-':
                    rev = True
                    o = o[1:]
                if o == 'pk':
                    o = qset.model._meta.pk
                else:
                    o = qset.model._meta.get_field(o) # raises FieldDoesNotExist
                ordering.append((o, rev))
                if o.unique:
                    break
        elif type(qset) == RawQuery:
            for o in qset.sortlist():
                rev = False
                if o[0] == '-':
                    rev = True
                    o = o[1:]
                ordering.append((o, rev))
                if o in qset.uniques:
                    break
    except FieldDoesNotExist:
        ordering = None
        pass
    # print "s5 {}".format(time.time() - start_time)
    # parse supplied paging info
    do_reverse = False
    first_page = start == 0
    last_page = False
    if ordering and request:
        try:
            try:
                page = json.loads(base64.b64decode(request.DATA.get('page', page)))
                if not page.get('_dir') in ('+', '-'):
                    raise FieldDoesNotExist
            except (ValueError, TypeError):
                raise FieldDoesNotExist

            qfilter = None
            qfilter2 = {}
            rfilter = []
            rfilter2 = []
            for (o, rev) in ordering:
                name = getattr(o, 'name', o)

                if not name in page:
                    raise FieldDoesNotExist

                if isinstance(page[name], dict) and "datetime" in page[name]:
                    data = datetime.datetime.fromtimestamp(page[name]["datetime"][0])
                    data = data.replace(microsecond=page[name]["datetime"][1])
                elif isinstance(page[name], dict) and "date" in page[name]:
                    data = datetime.date.fromtimestamp(page[name]["date"][0])
                else:
                    data = page[name]

                if type(qset) == QuerySet:
                    if rev == (page['_dir'] == '+'):
                        att = name + "__lt"
                    else:
                        att = name + "__gt"

                    qfilter3 = qfilter2.copy()
                    qfilter3[att] = data
                    if not qfilter:
                        qfilter = Q(**qfilter3)
                    else:
                        qfilter = qfilter | Q(**qfilter3)

                    if o.unique:
                        break
                    qfilter2[o.name] = data
                elif type(qset) == RawQuery:
                    if rev == (page['_dir'] == '+'):
                        att = name + "<"
                    else:
                        att = name + ">"

                    data = qset.serialize(data)
                    rfilter3 = list(rfilter2) + [att + data]
                    rfilter.append("(" + ") AND (".join(rfilter3) + ")")

                    if name in qset.uniques:
                        break
                    rfilter2.append(name + "=" + data)

            if type(qset) == QuerySet:
                qset = qset.filter(qfilter)
            elif type(qset) == RawQuery:
                qset.addWhere("((" + ") OR (".join(rfilter) + "))")
            if page['_dir'] == '-':
                do_reverse = True
            del ret['start']
            first_page = False
        except FieldDoesNotExist:
            pass
    # print "s6 {}".format(time.time() - start_time)
    if do_reverse:
        if type(qset) == QuerySet:
            if size:
                qset = list(qset.reverse()[:size+1])
                if len(qset) < size+1:
                    first_page = True
                else:
                    qset = qset[1:]
            qset.reverse()
        elif type(qset) == RawQuery:
            qset.reverse()
    # print "s7 {}".format(time.time() - start_time)
    if type(qset) == RawQuery:
        qset.offset = start
        if size:
            qset.limit = size+1
    elif size and qset != None:
        if do_reverse:
            qset = qset[start:start+size]
        else:
            t8 = time.time()
            # qset = list(qset[start:start+size+1])
            qset = qset[start:start+size+1]

            if qset.count() <= size:
                last_page = True
            else:
                qset = qset[:size]
            # print "t8 {}".format(time.time() - t8)
    # print "s8 {}".format(time.time() - start_time)
    if type(qset) == RawQuery:
        qset = list(qset.run())
        if do_reverse:
            if size:
                if qset.count() <= size:
                    first_page = True
                else:
                    qset = qset[:-1]
            qset.reverse()
        elif size:
            if qset.count() <= size:
                last_page = True
            else:
                qset = qset[:-1]
    # print "s9 {}".format(time.time() - start_time)
    # get list of data
    if "key" in extra and "value" in extra and len(extra) == 2 and "*" in exclude:
        ret['data'] = {}
        for qr in qset:
            qr = todata(qr)
            if hasattr(qr.key, '__func__'):
                key = qr.key()
            else:
                key = qr.key
            if hasattr(qr.value, '__func__'):
                value = qr.value()
            else:
                value = qr.value
            ret['data'][key] = rest_serialize(value)
    else:
        ret['data'] = []
        if qset:
            # # print fields
            for qr in qset:
                if type(qr) in [float, int, int, str, Decimal]:
                    ret['data'].append(qr)
                else:
                    # FIXME... wrong graph data????
                    ret['data'].append(restGet(request, todata(qr), model=model, fields=fields, extra=extra, exclude=exclude, fkey_depth=fkey_depth, recurse_into=recurse_into, filter=filter, ignore_noattr=ignore_noattr, accept_list=['data'], orig_request=orig_request, orig_objs=orig_objs))
    # print "s10 {}".format(time.time() - start_time)
    if accept_list and 'data' in accept_list:
        return restResult(request, ret['data'], accept_list)
    # print "s11 {}".format(time.time() - start_time)
    # fill in next/previous
    if ordering:
        try:
            if not first_page:
                ret['page_prev'] = _order_list(ordering, qset[0], '-')
            else:
                ret['page_before'] = _order_list(ordering, qset[0], '-')
        except IndexError:
            pass
        try:
            if do_reverse:
                offset = -1
            else:
                offset = size-1
            if size > 0 and not last_page:
                ret['page_next'] = _order_list(ordering, qset[offset], '+')
            elif size > 0:
                ret['page_after'] = _order_list(ordering, qset[offset], '+')
        except IndexError:
            try:
                ret['page_after'] = _order_list(ordering, qset[-1], '+')
            except:
                pass
    else:
        if not first_page:
            first = start - size
            if first < 0:
                first = 0
            ret['page_prev'] = _order_list(None, {'start': first}, '=')
        if size > 0 and not last_page and not ('count' in ret and start+size > ret['count']):
            ret['page_next'] = _order_list(None, {'start': start + size}, '=')
    # print "s12 {}".format(time.time() - start_time)
    if return_httpresponse:
        return restResult(request, ret, accept_list)
    return ret["data"]

def restSet(request, obj, initial=None, model=None, fields=None, form=None, data_override={}, retdata_func=None, retdata_args=[], return_object=False, accept_list=None, formopts={}, request_override={}):
    """
    request: HttpRequest object
    obj: Object to change
    initial: (optional) dictionary of default values, can be overwritten in http request
    model: (optional) model of the queryset
    fields: (optional) fields that can be set, default to all fields
    form: (optional) form used to validate change, defaults to no validation
    data_override: (optional) override object data after form processing
    retdata_func: (optional) function to return extra data (in dict form) for successful return status
    retdata_args: (optional) options to pass to retdata_func
    return_object: (optional, default False) return additional db object
    accept_list: (optional) requested return format
    RETURN: HttpResponse

    process a RESTful POST/PUT request to change an objects.  User is responsible for checking permissions, etc before calling.
    """
    # get valid fields from form
    if form:
        if not fields:
            fields = form.base_fields

    # filter input based on fields
    if not fields:
        fields = _getFields(obj, model)
        try:
            fields.remove('id')
        except ValueError:
            pass

    data = request.DATA

    # validate input
    if form:
        ldict = MultiValueDict({"__request": [request]})
        for v in request_override:
            ldict[v] = request_override[v]
        if obj:
            newform = form(instance=obj, initial=initial, **formopts)
            for f in newform.initial:
                if not (f in request.DATA or ('_clear_' + f) in request.DATA):
                    if isinstance(newform.initial[f], list):
                        ldict.setlist(f, newform.initial[f])
                    else:
                        ldict[f] = newform.initial[f]
        data = helpers.mergeDicts(ldict, request.POST, request.GET)

        if obj:
            form = form(data, files=request.FILES, instance=obj, initial=initial, **formopts)
        else:
            form = form(data, files=request.FILES, initial=initial, **formopts)
        if form.is_valid():
            if data_override:
                obj = form.save(commit=False)
            else:
                obj = form.save()
        else:
            errs = {}
            errs_all = ""
            for f in form.fields:
                if form[f].errors:
                    errs[f] = []
                    for e in form[f].errors:
                        errs[f].append(e)

            for e in form.non_field_errors():
                if not errs.get('', None):
                    errs[''] = []
                errs[''].append(e)
            for f in errs:
                errs_all += "%s: %s\n" % ( f, ", ".join(errs[f]) )

            ret = restResult(request, {'status': False, 'error': errs_all, 'errors': errs}, accept_list=accept_list)
            if return_object:
                return (ret, None,)
            else:
                return ret
    else:
        if not obj:
            obj = model()
            for f in initial:
                setattr(obj, f, initial[f])
        for f in fields:
            if f in request_override:
                setattr(obj, f, request_override[f])
            if f in data:
                setattr(obj, f, data[f])
        if not data_override:
            obj.save()

    # apply override
    if data_override:
        for k in data_override:
            setattr(obj, k, data_override[k])
        obj.save()
        if form:
            form.save_m2m()

    if hasattr(obj, 'pk'):
        retdata = {'status': True, 'id': obj.pk}
    else:
        retdata = {'status': True}
    if retdata_func:
        retdata['data'] = retdata_func(*retdata_args, obj=obj)
    ret = restResult(request, retdata, accept_list=accept_list)
    if return_object:
        return (ret, obj,)

    return ret

def restAdd(request, *args, **kwargs):
    """
    request: HttpRequest object
    other options: same as restSet

    RETURN: HttpResponse

    process a RESTful POST/PUT request to add an objects.  User is responsible for checking permissions, etc before calling.
    """
    return restSet(request, None, *args, **kwargs)

def restDelete(request, obj, accept_list=None):
    """
    request: HttpRequest object
    obj: Object to delete
    accept_list: (optional) requested return format
    RETURN: HttpResponse

    process a RESTful DELETE request to delete an objects.  User is responsible for checking permissions, etc before calling.
    """
    try:
        oldid = obj.pk
        obj.delete()
        return restResult(request, {'status': True, 'id': oldid}, accept_list=accept_list)
    except:
        return restResult(request, {'status': False, 'error': 'Cannot delete'}, accept_list=accept_list)

def restPermissionDenied(request, error="permission denied", error_code=403, status=403,
                         component=None, component_id=None):
    return restResult(request, dict(status=False, error=error, error_code=error_code), status=status)

def restNotFound(request):
    return restStatus(request, False, error="resource not found", error_code=404)

def restOK(request):
    return restStatus(request, True)

def restError(request, error, error_code=None):
    return restStatus(request, False, error=error, error_code=error_code)

def restSuccessNoContent(request):
    return HttpResponse(status=204)

def restStatus(request, status, data={}, accept_list=None, **kwargs):
    """
    request: HttpRequest object
    status: return status value
    accept_list: (optional) requested return format
    data: any other values to be returned
    kwargs: any other values to be returned
    RETURN: HttpResponse

    return a status using REST framework
    """
    if type(data) in [str, str]:
        if not status:
            data = {"error": data}
        else:
            data = {"message": data}
    ret = data.copy()
    ret.update(kwargs)
    ret['status'] = status
    # if settings.DEBUG and "error" in kwargs:
    #   print "RPC ERROR: {0}".format(kwargs["error"])
    return restResult(request, ret, accept_list)

_chunk_badchars = ''.join(list(chr(x) for x in list(range(0, ord('0'))) + list(range(ord('9')+1, ord('A'))) + list(range(ord('Z')+1, ord('a'))) + list(range(ord('z')+1, 256)) ))

def chunkUploadedFile(request, session):
    if (request.user and request.user.is_authenticated):
        uid = str(request.user.pk)
    else:
        uid = 'X'

    session = str(session).translate(None, _chunk_badchars)

    datafile = os.path.join(CHUNKDIR, uid + '-' + session)

    f = open(datafile, "r")
    data = eval(f.read())
    f.close()

    f = open(data['path'], "r")

    return UploadedFile(file=f, name=data.get('name'), content_type=data.get('content_type'), size=data.get('size'))

def chunkUploadView(request):
    for k in request.DATA:
        if k[-5:] == '.path':
            vname = k[:-5]
            break
    else:
        raise ValueError('No file uploaded')

    if not os.path.isdir(CHUNKDIR):
        os.makedirs(CHUNKDIR);

    if (request.user and request.user.is_authenticated):
        uid = str(request.user.pk)
    else:
        uid = 'X'

    infile = request.DATA[vname + '.path']
    name = request.DATA.get(vname + '.name', 'noname')
    if not name:
        name = 'noname'
    size = int(request.DATA[vname + '.size'])
    content_type = request.DATA.get(vname + '.content_type')

    sid = str(os.path.basename(infile))
    sid = sid.translate(None, _chunk_badchars)

    datafile = os.path.join(CHUNKDIR, uid + '-' + sid)
    outfile = os.path.join(datafile + "_file" + os.path.splitext(name)[1])

    if os.path.isfile(datafile):
        raise ValueError("Session ID Conflict: %s" % sid)

    os.rename(infile, outfile)

    f = open(datafile + ".tmp", "w")
    f.write(repr({
        'name': name,
        'size': size,
        'content_type': content_type,
        'path': outfile,
        'created': datetime.datetime.now(),
    }))
    f.close()
    os.rename(datafile + ".tmp", datafile)

    return restStatus(request, True, {'session-id': sid, 'name': name, 'content_type': content_type, 'size': size})


def restFlat(request, qset, fields, name, size=10000, header_cols=None):
    return HttpResponse(prettyJSON(list(qset.values_list(*fields))), content_type="application/json", status=200)


def restCSV(request, qset, fields, name, size=10000, header_cols=None):
    if size > 1200 and qset.count() > 1200:
        # large data set lets stream
        qset = qset[:size]
        response = rcsv.generateCSVStream(qset, fields, name)
        response['Cache-Control'] = 'no-cache'
        response['Content-Disposition'] = 'attachment; filename="{0}"'.format(name)
    else:
        response = HttpResponse(content_type='text/csv')
        response['Content-Disposition'] = 'attachment; filename="{0}"'.format(name)
        qset = qset[:size]
        rcsv.generateCSV(qset, fields, name, output=response, header_cols=header_cols)
    return response


def restExcel(request, qset, fields, name, size=10000, **kwargs):
    return rxsl.qsetToExcel(request, qset[:size], fields, name)


def restJSON(request, qset, fields, name, size=10000):
    data = restList(None, qset, fields=fields, size=size, accept_list=["application/json"], return_httpresponse=False)
    response = HttpResponse(prettyJSON(data), content_type="application/json")
    response['Content-Disposition'] = 'attachment; filename="{0}"'.format(name)
    return response




from rest import decorators as rd
from rest import views as rv
from rest import helpers as rh
from account import models as am
from objict import objict


@rd.url(r'^member/device$')
@rd.url(r'^member/device/(?P<pk>\d+)$')
@rd.login_required
def rest_on_member_device(request, pk=None):
    return am.MemberDevice.on_rest_request(request, pk)


@rd.urlPOST(r'^member/device/register$')
@rd.login_required
def rest_on_member_device_register(request):
    # this requires a JWT with a device_id
    if not request.device_id:
        return rv.restPermissionDenied(request, "requires jwt with device id")
    cm_token = request.DATA.get("cm_token")
    device_id = request.device_id
    if not device_id:
        return rv.restStatus(request, False, error="missing device id")
    md = am.MemberDevice.objects.filter(uuid=device_id).last()
    if md is None:
        md = am.MemberDevice.createFromRequest(request, uuid=device_id, user=request.member)
    elif cm_token is not None and md.cm_token != cm_token:
        md.cm_token = cm_token
        md.cm_provider = "fcm"
        md.save()
    return rv.restStatus(request, True)


@rd.urlPOST(r'^member/device/register$')
@rd.login_required
def rest_on_member_device_register(request):
    # this requires a JWT with a device_id
    if not request.device_id:
        return rv.restPermissionDenied(request, "requires jwt with device id")

    md = am.MemberDevice.register(request, request.member, request.device_id)
    return rv.restStatus(request, md is not None)


@rd.urlPOST(r'^member/device/unregister$')
@rd.login_required
def rest_on_member_device_unregister(request):
    # this requires a JWT with a device_id
    if not request.device_id:
        return rv.restPermissionDenied(request, "requires jwt with device id")
    return rv.restStatus(request, am.MemberDevice.unregister(request.member, request.device_id))


@rd.urlPOST(r'^member/device/notify$')
@rd.login_required
def rest_on_member_device_notify(request):
    device_id, md = getMemberDevice(request)
    payload = request.DATA.get("payload")
    if payload is None:
        return rv.restPermissionDenied(request, error="missing payload")
    if md is None:
        return rv.restPermissionDenied(request, error=F"missing device: {device_id}")
    if payload.kind == "notification":
        r = md.sendNotification(payload.title, payload.body)
    else:
        r = md.sendData(payload)
    if r.status_code != 200:
        rh.log_error(payload, r.reason, r.status_code, r.text)
        return rv.restStatus(request, False, error=r.reason, code=r.status_code)
    return rv.restGet(request, r.json())


def getMemberDevice(request):
    device_id = request.DATA.get(["device_id", "deviceID", "device"])
    if not device_id:
        return None, None
    return device_id, am.MemberDevice.objects.filter(uuid=device_id).last()




_run = []
_cleanup = []

def register_run(func):
    global _run
    if not func in _run:
        _run.append(func)

def register_cleanup(func):
    global _cleanup
    if not func in _cleanup:
        _cleanup.append(func)

def run():
    global _run
    global _cleanup
    for f in _run:
        f()
    for f in _cleanup:
        f()
    _run = []
    _cleanup = []
